#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import TYPE_CHECKING

import hopsworks
from fastmcp import (
    Context,  # noqa: TC002
)
from hopsworks.mcp.models.feature_group import Feature, FeatureGroup
from hopsworks.mcp.utils.tags import TAGS


if TYPE_CHECKING:
    from fastmcp import (
        FastMCP,
    )


class FeatureGroupTools:
    """Tools for managing feature groups in Hopsworks MCP."""

    def __init__(self, mcp: FastMCP):
        self.mcp = mcp
        self.mcp.tool(tags=[TAGS.FEATURE_GROUP, TAGS.READ, TAGS.STATEFUL])(
            self.get_feature_groups
        )
        self.mcp.tool(tags=[TAGS.FEATURE_GROUP, TAGS.READ, TAGS.STATEFUL])(
            self.get_feature_group_versions
        )
        self.mcp.tool(tags=[TAGS.FEATURE_GROUP, TAGS.READ, TAGS.STATEFUL])(
            self.get_feature_group_details
        )
        self.mcp.tool(tags=[TAGS.FEATURE_GROUP, TAGS.READ, TAGS.STATEFUL])(
            self.get_feature_group_details
        )
        self.mcp.tool(tags=[TAGS.FEATURE_GROUP, TAGS.READ, TAGS.STATEFUL])(
            self.preview_feature_group
        )
        self.mcp.tool(tags=[TAGS.FEATURE_GROUP, TAGS.READ, TAGS.STATEFUL])(
            self.get_features
        )

    def _get_feature_group_versions(self, name: str | None = None):
        # Get the current project and its feature groups
        try:
            project = hopsworks.get_current_project()
        except hopsworks.ProjectException:
            raise RuntimeError(
                "No active Hopsworks project found, use login tool."
            ) from None

        return project.get_feature_store().get_feature_groups(name=name)

    def _get_feature_group_version(
        self, name: str | None = None, version: int | None = None
    ):
        fgs = self._get_feature_group_versions(name)
        if version is not None:
            for fg in fgs:
                if fg.version == version:
                    return fg
            raise RuntimeError(f"Feature group {name} v{version} not found.")
        try:
            return sorted(fgs, key=lambda fg: fg.version, reverse=True)[0]
        except IndexError:
            raise RuntimeError(f"Feature group {name} not found.") from None

    async def get_feature_groups(  # docsig: disable
        self, ctx: Context
    ) -> list[FeatureGroup]:
        """Get the latest versions of all feature groups in the project."""
        await ctx.info("Retrieving feature groups...")

        fgs = self._get_feature_group_versions()
        fg_names = {fg.name for fg in fgs}
        fg_latest_version = {
            name: sorted([fg.version for fg in fgs if fg.name == name], reverse=True)[0]
            for name in fg_names
        }
        return sorted(
            [
                FeatureGroup(id=fg.id, name=fg.name, version=fg.version)
                for fg in fgs
                if fg.version == fg_latest_version[fg.name]
            ],
            key=lambda fg: (fg.name, fg.version),
        )

    async def get_feature_group_versions(  # docsig: disable
        self, ctx: Context, name: str
    ) -> list[int]:
        """Get all versions of a feature group with the specified name."""
        await ctx.info("Retrieving feature groups...")

        fgs = self._get_feature_group_versions(name)
        return sorted([fg.version for fg in fgs])

    async def get_feature_group_details(  # docsig: disable
        self,
        ctx: Context,
        name: str,
        version: int | None = None,
    ) -> FeatureGroup:
        """Get the detailed description of a feature group with the specified name and version (latest by default)."""
        await ctx.info(
            f"Retrieving details of {name}{f' v{version}' if version else ''} feature group..."
        )

        fg = self._get_feature_group_version(name, version)
        return FeatureGroup(
            id=fg.id,
            name=fg.name,
            version=fg.version,
            description=fg.description,
            location=fg.location,
            event_time=fg.event_time,
            online_enabled=fg.online_enabled,
            topic_name=fg.topic_name,
            notification_topic_name=fg.notification_topic_name,
            deprecated=fg.deprecated,
        )

    async def preview_feature_group(  # docsig: disable
        self,
        ctx: Context,
        name: str,
        version: int | None = None,
        n: int = 10,
    ) -> dict[str, list[str | int | float | None]]:
        """Preview the first n (10 by default) rows of a feature group with the specified name and version (latest by default).

        The tool can be useful to figure out the actual schema of the feature group in case the feature metadata is incomplete or confusing.
        """
        # TODO: the function is partially complete, we should add a method to list data in columnar format to the API and use it here instead.
        await ctx.info(
            f"Retrieving preview of {name}{f' v{version}' if version else ''} feature group..."
        )

        fg = self._get_feature_group_version(name, version)
        preview = fg.show(n, fg.online_enabled)

        try:
            import pandas as pd

            if isinstance(preview, pd.DataFrame):
                return {
                    str(k): [
                        x if isinstance(x, (int, float, type(None))) else str(x)
                        for x in vs
                    ]
                    for k, vs in preview.to_dict(orient="list").items()
                }
        except ImportError:
            pass

        try:
            import polars as pl

            if isinstance(preview, pl.DataFrame):
                return {k: vs.to_list() for k, vs in preview.to_dict().items()}
        except ImportError:
            pass

        raise RuntimeError(
            f"Unable to convert preview to dictionary. Here's the raw preview:\n{preview}"
        )

    async def get_features(  # docsig: disable
        self,
        ctx: Context,
        name: str,
        version: int | None = None,
    ) -> list[Feature]:
        """Get the features of a feature group with the specified name and version (latest by default)."""
        await ctx.info(
            f"Retrieving features of {name}{f' v{version}' if version else ''} feature group..."
        )

        fg = self._get_feature_group_version(name, version)
        return sorted(
            [
                Feature(
                    name=f.name,
                    type=f.type,
                    description=f.description,
                    primary=f.primary,
                    event_time=fg.event_time == f.name,
                )
                for f in fg.features
            ],
            key=lambda feature: feature.name,
        )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import TYPE_CHECKING

from hopsworks_common import client
from hsfs.core import data_source as ds
from hsfs.core import data_source_data as dsd


if TYPE_CHECKING:
    from hsfs import storage_connector as sc


class DataSourceApi:
    def get_databases(self, storage_connector: sc.StorageConnector) -> list[str]:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            storage_connector._featurestore_id,
            "storageconnectors",
            storage_connector._name,
            "data_source",
            "databases",
        ]

        return _client._send_request("GET", path_params)

    def get_crm_resources(
        self, storage_connector: sc.StorageConnector
    ) -> dsd.DataSourceData:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            storage_connector._featurestore_id,
            "storageconnectors",
            storage_connector._name,
            "data_source",
            "resources",
        ]
        return dsd.DataSourceData.from_response_json(
            _client._send_request("GET", path_params)
        )

    def get_tables(
        self, storage_connector: sc.StorageConnector, database: str
    ) -> list[ds.DataSource]:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            storage_connector._featurestore_id,
            "storageconnectors",
            storage_connector._name,
            "data_source",
            "tables",
        ]

        query_params = {"database": database}

        return ds.DataSource.from_response_json(
            _client._send_request("GET", path_params, query_params),
            storage_connector=storage_connector,
        )

    def get_no_sql_data(
        self,
        storage_connector: sc.StorageConnector,
        data_source: ds.DataSource,
        use_cached=True,
    ) -> dsd.DataSourceData:
        if storage_connector.type == "REST":
            return self._get_rest_data(storage_connector, data_source, use_cached)
        if storage_connector.type == "CRM":
            return self._get_crm_data(storage_connector, data_source, use_cached)
        raise ValueError("This connector type does not support fetching NoSQL data.")

    def _get_rest_data(
        self,
        storage_connector: sc.StorageConnector,
        data_source: ds.DataSource,
        use_cached=True,
    ) -> dsd.DataSourceData:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            storage_connector._featurestore_id,
            "storageconnectors",
            storage_connector._name,
            "data_source",
            "resources",
            data_source.table,
        ]

        query_params = {"forceRefetch": not use_cached}
        return dsd.DataSourceData.from_response_json(
            _client._send_request(
                "POST",
                path_params,
                query_params=query_params,
                headers={"content-type": "application/json"},
                data=json.dumps(data_source.rest_endpoint.to_dict()),
            )
        )

    def _get_crm_data(
        self,
        storage_connector: sc.StorageConnector,
        data_source: ds.DataSource,
        use_cached=True,
    ) -> dsd.DataSourceData:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            storage_connector._featurestore_id,
            "storageconnectors",
            storage_connector._name,
            "data_source",
            "resources",
            data_source.table,
        ]
        query_params = data_source.to_dict()
        if use_cached is not None:
            query_params["forceRefetch"] = not use_cached

        return dsd.DataSourceData.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )

    def get_data(self, data_source: ds.DataSource) -> dsd.DataSourceData:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            data_source._storage_connector._featurestore_id,
            "storageconnectors",
            data_source._storage_connector._name,
            "data_source",
            "data",
        ]

        query_params = data_source.to_dict()

        return dsd.DataSourceData.from_response_json(
            _client._send_request("GET", path_params, query_params)
        )

    def get_metadata(self, data_source: ds.DataSource) -> dict:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            data_source._storage_connector._featurestore_id,
            "storageconnectors",
            data_source._storage_connector._name,
            "data_source",
            "metadata",
        ]

        query_params = data_source.to_dict()

        return _client._send_request("GET", path_params, query_params)



#
#   Copyright 2026 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import pytest
from hopsworks_common.constants import PREDICTOR, SCALING_CONFIG
from hsml.scaling_config import (
    PredictorScalingConfig,
    ScaleMetric,
    TransformerScalingConfig,
)


class TestScalingConfig:
    def test_scale_metric_has_value(self):
        assert ScaleMetric.has_value("CONCURRENCY")
        assert ScaleMetric.has_value("RPS")
        assert not ScaleMetric.has_value("BOGUS")

    def test_predictor_scaling_config_accepts_scale_metric_string(self):
        sc = PredictorScalingConfig(min_instances=1, scale_metric="rps")
        assert sc.scale_metric == ScaleMetric.RPS

    def test_predictor_scaling_config_invalid_scale_metric_string(self):
        with pytest.raises(ValueError) as exc_info:
            PredictorScalingConfig(min_instances=1, scale_metric="bogus")

        assert "Invalid scale_metric" in str(exc_info.value)

    def test_predictor_scaling_config_invalid_scale_metric_type(self):
        with pytest.raises(ValueError) as exc_info:
            PredictorScalingConfig(min_instances=1, scale_metric=123)

        assert "scale_metric must be a string or ScaleMetric" in str(exc_info.value)

    def test_get_default_scaling_configuration_kserve_scale_to_zero_required(
        self, mocker
    ):
        mocker.patch(
            "hopsworks_common.client.is_scale_to_zero_required", return_value=True
        )

        sc = PredictorScalingConfig.get_default_scaling_configuration(
            PREDICTOR.SERVING_TOOL_KSERVE, None
        )

        assert sc.min_instances == 0
        assert sc.scale_metric.value == SCALING_CONFIG.SCALE_METRIC_CONCURRENCY
        assert sc.target == SCALING_CONFIG.DEFAULT_CONCURRENCY_TARGET
        assert (
            sc.panic_window_percentage == SCALING_CONFIG.DEFAULT_PANIC_WINDOW_PERCENTAGE
        )
        assert (
            sc.panic_threshold_percentage
            == SCALING_CONFIG.DEFAULT_PANIC_THRESHOLD_PERCENTAGE
        )
        assert sc.stable_window_seconds == SCALING_CONFIG.DEFAULT_STABLE_WINDOW_SECONDS
        assert (
            sc.scale_to_zero_retention_seconds
            == SCALING_CONFIG.DEFAULT_SCALE_TO_ZERO_RETENTION_SECONDS
        )

    def test_get_default_scaling_configuration_transformer_type(self, mocker):
        mocker.patch(
            "hopsworks_common.client.is_scale_to_zero_required", return_value=False
        )

        sc = PredictorScalingConfig.get_default_scaling_configuration(
            PREDICTOR.SERVING_TOOL_DEFAULT, 1, component_type="transformer"
        )

        assert isinstance(sc, TransformerScalingConfig)
        assert sc.min_instances == 1

    def test_get_default_scaling_configuration_non_kserve_min_zero_raises(self, mocker):
        mocker.patch(
            "hopsworks_common.client.is_scale_to_zero_required", return_value=False
        )

        with pytest.raises(ValueError) as exc_info:
            PredictorScalingConfig.get_default_scaling_configuration(
                PREDICTOR.SERVING_TOOL_DEFAULT, 0
            )

        assert "Minimum number of instances cannot be 0" in str(exc_info.value)

    def test_get_default_scaling_configuration_kserve_requires_zero(self, mocker):
        mocker.patch(
            "hopsworks_common.client.is_scale_to_zero_required", return_value=True
        )

        with pytest.raises(ValueError) as exc_info:
            PredictorScalingConfig.get_default_scaling_configuration(
                PREDICTOR.SERVING_TOOL_KSERVE, 1
            )

        assert "Scale-to-zero is required" in str(exc_info.value)

    def test_from_json_to_json_roundtrip(self):
        json_payload = {
            "predictor_scaling_config": {
                "min_instances": 1,
                "max_instances": 2,
                "scale_metric": "CONCURRENCY",
                "target": 10,
                "panic_window_percentage": 5.0,
                "panic_threshold_percentage": 150.0,
                "stable_window_seconds": 30,
                "scale_to_zero_retention_seconds": 60,
            }
        }

        sc = PredictorScalingConfig.from_json(json_payload)

        assert sc.min_instances == 1
        assert sc.max_instances == 2
        assert sc.scale_metric.value == "CONCURRENCY"
        assert sc.target == 10
        assert sc.panic_window_percentage == 5.0
        assert sc.panic_threshold_percentage == 150.0
        assert sc.stable_window_seconds == 30
        assert sc.scale_to_zero_retention_seconds == 60

        assert sc.to_json() == {
            "min_instances": 1,
            "scale_metric": "CONCURRENCY",
            "target": 10,
            "max_instances": 2,
            "panic_window_percentage": 5.0,
            "panic_threshold_percentage": 150.0,
            "stable_window_seconds": 30,
            "scale_to_zero_retention_seconds": 60,
        }

    def test_to_dict_camelizes_scaling_config(self):
        sc = PredictorScalingConfig(
            min_instances=1,
            max_instances=2,
            scale_metric="CONCURRENCY",
            target=10,
            panic_window_percentage=5.0,
            panic_threshold_percentage=150.0,
            stable_window_seconds=30,
            scale_to_zero_retention_seconds=60,
        )

        assert sc.to_dict() == {
            "predictorScalingConfig": {
                "minInstances": 1,
                "scaleMetric": "CONCURRENCY",
                "target": 10,
                "maxInstances": 2,
                "panicWindowPercentage": 5.0,
                "panicThresholdPercentage": 150.0,
                "stableWindowSeconds": 30,
                "scaleToZeroRetentionSeconds": 60,
            }
        }

    def test_from_json_missing_min_instances_raises(self):
        json_payload = {"predictor_scaling_config": {"scale_metric": "CONCURRENCY"}}

        with pytest.raises(ValueError) as exc_info:
            PredictorScalingConfig.from_json(json_payload)

        assert "missing 'min_instances'" in str(exc_info.value)



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
from typing import Literal

import humps
from hopsworks_apigen import public
from hopsworks_common import util


@public("hopsworks.git_file_status.GitFileStatus")
class GitFileStatus:
    def __init__(
        self,
        file=None,
        status=None,
        extra=None,
        **kwargs,
    ):
        self._file = file
        self._status = status
        self._extra = extra

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" in json_decamelized:
            if json_decamelized["count"] == 0:
                return []
            return [cls(**file_status) for file_status in json_decamelized["items"]]
        return cls(**json_decamelized)

    @public
    @property
    def file(self) -> str:
        """Path to the file."""
        return self._file

    @public
    @property
    def status(self) -> Literal[" ", "?", "M", "A", "D", "R", "C", "U"]:
        """Status of the file.

        The status is represented by a single character, [according](https://git-scm.com/docs/git-status#_output) to `git status`:

        | Unmodified           | `' '` |
        | Untracked            | `'?'` |
        | Modified             | `'M'` |
        | Added                | `'A'` |
        | Deleted              | `'D'` |
        | Renamed              | `'R'` |
        | Copied               | `'C'` |
        | Updated but unmerged | `'U'` |
        """
        return self._status

    @public
    @property
    def extra(self):
        """Extra contains additional information, such as the previous name in a rename."""
        return self._extra

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"GitFileStatus({self._file!r}, {self._status!r}, {self._extra!r})"



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from hopsworks_apigen import public
from hopsworks_common import client


if TYPE_CHECKING:
    from hsfs.feature_group import FeatureGroup
    from hsfs.feature_view import FeatureView
    from hsfs.training_dataset import TrainingDataset


@public("hopsworks.core.search_api.Project")
class Project:
    """Represents a project associated with a search result."""

    def __init__(self, project_id: int, project_name: str):
        self._log = logging.getLogger(__name__)
        self._id = project_id
        self._name = project_name

    @public
    @property
    def id(self) -> int:
        """Project ID."""
        return self._id

    @public
    @property
    def name(self) -> str:
        """Project name."""
        return self._name

    def json(self) -> dict:
        """Convert to JSON-serializable dictionary.

        Returns:
            A dictionary representation of the project.
        """
        return {"id": self._id, "name": self._name}

    def __repr__(self):
        return f"Project(id={self._id}, name='{self._name}')"


@public("hopsworks.core.search_api.Highlights")
class Highlights:
    """Container for search result highlights showing where matches occurred.

    The results are highlighted by wrapping the matched terms in `<em>` tags.
    Check the [OpenSearch Highlight Queries](https://docs.opensearch.org/latest/search-plugins/searching-data/highlight/) for more details.
    """

    def __init__(self, highlights_data: dict):
        self._log = logging.getLogger(__name__)
        self._raw_data = highlights_data
        self._name = highlights_data.get("name")
        self._description = highlights_data.get("description")
        self._tags = highlights_data.get("tags", [])
        self._keywords = highlights_data.get("keywords", [])
        self._features = highlights_data.get("features", [])
        self._source_feature_groups = highlights_data.get("sourceFeatureGroups", [])

    @public
    @property
    def name(self) -> str | None:
        """Highlighted name with the matched parts enwrapped in `<em>` tags."""
        return self._name

    @public
    @property
    def description(self) -> str | None:
        """Highlighted description with the matched parts enwrapped in `<em>` tags."""
        return self._description

    @public
    @property
    def tags(self) -> list:
        """List of highlighted tags with the matched parts enwrapped in `<em>` tags."""
        return self._tags

    @public
    @property
    def keywords(self) -> list:
        """Highlighted keywords with the matched parts enwrapped in `<em>` tags."""
        return self._keywords

    @public
    @property
    def features(self) -> list:
        """Highlighted features with the matched parts enwrapped in `<em>` tags."""
        return self._features

    @public
    @property
    def source_feature_groups(self) -> list:
        """Highlighted source feature groups with the matched parts enwrapped in `<em>` tags."""
        return self._source_feature_groups

    @property
    def raw_data(self) -> dict:
        """Raw highlights data."""
        return self._raw_data

    def has_highlights(self) -> bool:
        """Check if there are any highlights.

        Returns:
            Whether any of the highlight fields contain data.
        """
        return bool(
            self._name
            or self._description
            or self._tags
            or self._keywords
            or self._features
            or self._source_feature_groups
        )

    def json(self) -> dict:
        """Convert to JSON-serializable dictionary.

        Returns:
            A dictionary representation of the highlights.
        """
        return {
            "name": self._name,
            "description": self._description,
            "tags": self._tags,
            "keywords": self._keywords,
            "features": self._features,
            "source_feature_groups": self._source_feature_groups,
        }

    def __repr__(self):
        highlights = []
        if self._name:
            highlights.append("name")
        if self._description:
            highlights.append("description")
        if self._tags:
            highlights.append("tags")
        if self._keywords:
            highlights.append("keywords")
        if self._features:
            highlights.append("features")
        if self._source_feature_groups:
            highlights.append("source_feature_groups")

        if highlights:
            return f"Highlights({', '.join(highlights)})"
        return "Highlights(none)"


@public("hopsworks.core.search_api.SearchResultItem")
class SearchResultItem:
    """Base class for search result items."""

    def __init__(self, data: dict):
        self._log = logging.getLogger(__name__)
        self._href = data.get("href")
        self._name = data.get("name")
        self._version = data.get("version")
        self._description = data.get("description")
        self._highlights = Highlights(data.get("highlights", {}))
        self._raw_data = data

        # Extract project information
        project_id = data.get("parentProjectId")
        project_name = data.get("parentProjectName")
        self._project = (
            Project(project_id, project_name) if project_id and project_name else None
        )

    @public
    @property
    def href(self):
        """URL to get the full resource."""
        return self._href

    @public
    @property
    def name(self):
        """Name of the resource."""
        return self._name

    @public
    @property
    def version(self):
        """Version of the resource."""
        return self._version

    @public
    @property
    def description(self):
        """Description of the resource."""
        return self._description

    @public
    @property
    def highlights(self) -> Highlights:
        """Search highlights showing matched terms."""
        return self._highlights

    @public
    @property
    def project(self) -> Project | None:
        """Parent project of this resource."""
        return self._project

    @property
    def raw_data(self):
        """Raw data from the search result."""
        return self._raw_data

    def json(self) -> dict:
        """Convert to JSON-serializable dictionary.

        Returns:
            A dictionary representation of the search result item.
        """
        return {
            "href": self._href,
            "name": self._name,
            "version": self._version,
            "description": self._description,
            "highlights": self._highlights.json(),
            "project": self._project.json() if self._project else None,
        }

    def __repr__(self):
        version_str = f", version={self._version}" if self._version else ""
        if self._description:
            if len(self._description) > 50:
                description_preview = f"{self._description[:50]}..."
            else:
                description_preview = self._description
            description_str = f", description='{description_preview}'"
        else:
            description_str = ""
        return f"{self.__class__.__name__}(name='{self._name}'{version_str}{description_str}, project='{self._project}', highlights='{self._highlights}')"


@public("hopsworks.core.search_api.FeatureGroupSearchResult")
class FeatureGroupSearchResult(SearchResultItem):
    """Search result for a Feature Group."""

    @public
    def get(self) -> FeatureGroup | None:
        """Retrieve the full FeatureGroup object.

        This uses the project associated with this search result to obtain a
        connection to the feature store and then fetches the Feature Group
        with the given name and version.

        Returns:
            The full Feature Group object corresponding to this search result.

        Raises:
            Exception: If the connection to the feature store fails or the
                Feature Group cannot be retrieved.
        """
        fs = client.get_connection().get_feature_store(self.project.name)
        return fs.get_feature_group(self.name, version=self.version)


@public("hopsworks.core.search_api.FeatureViewSearchResult")
class FeatureViewSearchResult(SearchResultItem):
    """Search result for a Feature View."""

    @public
    def get(self) -> FeatureView | None:
        """Retrieve the full FeatureView object.

        This uses the project associated with this search result to obtain a
        connection to the feature store and then fetches the Feature View
        with the given name and version.

        Returns:
            The full FeatureView instance corresponding to this search result.

        Raises:
            Exception: If the connection to the feature store fails or the
                Feature View cannot be retrieved.
        """
        fs = client.get_connection().get_feature_store(self.project.name)
        return fs.get_feature_view(self.name, version=self.version)


@public("hopsworks.core.search_api.TrainingDatasetSearchResult")
class TrainingDatasetSearchResult(SearchResultItem):
    """Search result for a Training Dataset."""

    @public
    def get(self) -> TrainingDataset | None:
        """Retrieve the full TrainingDataset object.

        This uses the project associated with this search result to obtain a
        connection to the feature store and then fetches the Training Dataset
        with the given name and version.

        Returns:
            The full TrainingDataset instance corresponding to this search result.

        Raises:
            Exception: If the connection to the feature store fails or the
                Training Dataset cannot be retrieved.
        """
        fs = client.get_connection().get_feature_store(self.project.name)
        return fs.get_training_dataset(self.name, version=self.version)


@public("hopsworks.core.search_api.FeatureSearchResult")
class FeatureSearchResult(SearchResultItem):
    """Search result for a Feature."""


@public("hopsworks.core.search_api.FeaturestoreSearchResult")
class FeaturestoreSearchResult:
    """Container for all featurestore search results."""

    def __init__(self, response_data: dict):
        self._log = logging.getLogger(__name__)
        self._feature_groups = [
            FeatureGroupSearchResult(fg)
            for fg in response_data.get("featuregroups", [])
        ]
        self._feature_views = [
            FeatureViewSearchResult(fv) for fv in response_data.get("featureViews", [])
        ]
        self._training_datasets = [
            TrainingDatasetSearchResult(td)
            for td in response_data.get("trainingdatasets", [])
        ]
        self._features = [
            FeatureSearchResult(f) for f in response_data.get("features", [])
        ]

        # Store metadata about result counts
        self._feature_groups_offset = response_data.get("featuregroupsFrom", 0)
        self._feature_groups_total = response_data.get("featuregroupsTotal", 0)
        self._feature_views_offset = response_data.get("featureViewsFrom", 0)
        self._feature_views_total = response_data.get("featureViewsTotal", 0)
        self._training_datasets_offset = response_data.get("trainingdatasetsFrom", 0)
        self._training_datasets_total = response_data.get("trainingdatasetsTotal", 0)
        self._features_offset = response_data.get("featuresFrom", 0)
        self._features_total = response_data.get("featuresTotal", 0)

    @public
    @property
    def feature_groups(self) -> list[FeatureGroupSearchResult]:
        """List of Feature Group search results."""
        return self._feature_groups

    @public
    @property
    def feature_views(self) -> list[FeatureViewSearchResult]:
        """List of Feature View search results."""
        return self._feature_views

    @public
    @property
    def training_datasets(self) -> list[TrainingDatasetSearchResult]:
        """List of Training Dataset search results."""
        return self._training_datasets

    @public
    @property
    def features(self) -> list[FeatureSearchResult]:
        """List of Feature search results."""
        return self._features

    @public
    @property
    def feature_groups_offset(self) -> int:
        """Total offset for the return list of feature groups within the whole result."""
        return self._feature_groups_offset

    @public
    @property
    def feature_views_offset(self) -> int:
        """Total offset for the return list of feature views within the whole result."""
        return self._feature_views_offset

    @public
    @property
    def training_datasets_offset(self) -> int:
        """Total offset for the return list of training datasets within the whole result."""
        return self._training_datasets_offset

    @public
    @property
    def features_offset(self) -> int:
        """Total offset for the return list of features within the whole result."""
        return self._features_offset

    @public
    @property
    def feature_groups_total(self) -> int:
        """Total number of Feature Groups matching the search."""
        return self._feature_groups_total

    @public
    @property
    def feature_views_total(self) -> int:
        """Total number of Feature Views matching the search."""
        return self._feature_views_total

    @public
    @property
    def training_datasets_total(self) -> int:
        """Total number of Training Datasets matching the search."""
        return self._training_datasets_total

    @public
    @property
    def features_total(self) -> int:
        """Total number of Features matching the search."""
        return self._features_total

    def json(self) -> dict:
        """Convert to JSON-serializable dictionary.

        Returns:
            JSONDictionary representation of the object.
        """
        return {
            "featuregroups": [fg.json() for fg in self._feature_groups],
            "featuregroupsFrom": self._feature_groups_offset,
            "featuregroupsTotal": self._feature_groups_total,
            "featureviews": [fv.json() for fv in self._feature_views],
            "featureviewsFrom": self._feature_views_offset,
            "featureviewsTotal": self._feature_views_total,
            "trainingdatasets": [td.json() for td in self._training_datasets],
            "trainingdatasetsFrom": self._training_datasets_offset,
            "trainingdatasetsTotal": self._training_datasets_total,
            "features": [f.json() for f in self._features],
            "featuresFrom": self._features_offset,
            "featuresTotal": self._features_total,
        }

    def __repr__(self):
        return (
            f"FeaturestoreSearchResult("
            f"feature_groups={len(self._feature_groups)}/{self._feature_groups_total}, "
            f"feature_views={len(self._feature_views)}/{self._feature_views_total}, "
            f"training_datasets={len(self._training_datasets)}/{self._training_datasets_total}, "
            f"features={len(self._features)}/{self._features_total})"
        )



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from hopsworks_common import client
from hsfs import decorators, storage_connector


if TYPE_CHECKING:
    from hsfs.core.explicit_provenance import Links


class StorageConnectorApi:
    @decorators.catch_not_found(
        "hsfs.storage_connector.StorageConnector", fallback_return=None
    )
    def _get(self, feature_store_id: int, name: str) -> dict[str, Any]:
        """Returning response dict instead of initialized object."""
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            name,
        ]
        query_params = {"temporaryCredentials": True}

        return _client._send_request("GET", path_params, query_params=query_params)

    def get(
        self, feature_store_id: int, name: str
    ) -> storage_connector.StorageConnector:
        """Get storage connector with name.

        Parameters:
            feature_store_id: feature store id
            name: name of the storage connector

        Returns:
            the storage connector
        """
        storage_connector_json = self._get(feature_store_id, name)
        if storage_connector_json:
            return storage_connector.StorageConnector.from_response_json(
                storage_connector_json
            )
        return None

    def refetch(
        self, storage_connector_instance: storage_connector.StorageConnector
    ) -> storage_connector.StorageConnector:
        """Refetch the storage connector from Hopsworks, updating temporary credentials.

        Parameters:
            storage_connector_instance: The storage connector to refetch.

        Returns:
            The updated storage connector instance.
        """
        return storage_connector_instance.update_from_response_json(
            self._get(
                storage_connector_instance._featurestore_id,
                storage_connector_instance.name,
            )
        )

    def get_online_connector(
        self, feature_store_id: int
    ) -> storage_connector.OnlineStorageConnector:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            "onlinefeaturestore",
        ]

        return storage_connector.StorageConnector.from_response_json(
            _client._send_request("GET", path_params)
        )

    def get_kafka_connector(
        self, feature_store_id: int, external: bool = False
    ) -> storage_connector.KafkaConnector:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            "kafka_connector",
            "byok",
        ]
        query_params = {"external": external}

        return storage_connector.StorageConnector.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )

    def get_feature_groups_provenance(self, storage_connector_instance) -> Links:
        """Get the generated feature groups using this storage connector, based on explicit provenance.

        These feature groups can be accessible or inaccessible. Explicit
        provenance does not track deleted generated feature group links, so deleted
        will always be empty.
        For inaccessible feature groups, only a minimal information is returned.

        Parameters:
            storage_connector_instance: Metadata object of storage connector.

        Returns:
            The feature groups generated using this storage connector.
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            storage_connector_instance._featurestore_id,
            "storageconnectors",
            storage_connector_instance.name,
            "provenance",
            "links",
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 0,
            "downstreamLvls": 1,
        }
        links_json = _client._send_request("GET", path_params, query_params)

        from hsfs.core import explicit_provenance

        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.DOWNSTREAM,
            explicit_provenance.Links.Type.FEATURE_GROUP,
        )

    def get_training_datasets_provenance(self, storage_connector_instance) -> Links:
        """Get the generated training datasets using this storage connector, based on explicit provenance.

        These training datasets can be accessible or inaccessible.
        Explicit provenance does not track deleted generated training dataset links, so deleted will always be empty.
        For inaccessible training datasets, only a minimal information is returned.

        Parameters:
            storage_connector_instance: Metadata object of storage connector.

        Returns:
            The training datasets generated using this storage connector.
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            storage_connector_instance._featurestore_id,
            "storageconnectors",
            storage_connector_instance.name,
            "provenance",
            "links",
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 0,
            "downstreamLvls": 1,
        }
        links_json = _client._send_request("GET", path_params, query_params)

        from hsfs.core import explicit_provenance

        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.DOWNSTREAM,
            explicit_provenance.Links.Type.TRAINING_DATASET,
        )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

import humps
from hopsworks_apigen import public
from hopsworks_common import util
from hopsworks_common.core.rest_endpoint import RestEndpointConfig
from hsfs import storage_connector as sc


if TYPE_CHECKING:
    from hsfs import feature_group as fg
    from hsfs.core import data_source_data as dsd
    from hsfs.core.explicit_provenance import Links
    from hsfs.training_dataset import TrainingDataset


@public
class DataSource:
    """Metadata object used to provide data source information.

    You can obtain data sources using [`FeatureStore.get_data_source`][hsfs.feature_store.FeatureStore.get_data_source].

    The DataSource class encapsulates the details of a data source that can be used for reading or writing data.
    It supports various types of sources, such as SQL queries, database tables, file paths, and storage connectors.

    Parameters:
        query: SQL query string for the data source, if applicable.
        database: Name of the database containing the data source.
        group: Group or schema name for the data source.
        table: Table name for the data source.
        path: File system path for the data source.
        storage_connector: Storage connector object holds configuration for accessing the data source.
        metrics: List of metric column names for the data source.
        dimensions: List of dimension column names for the data source.
        rest_endpoint: REST endpoint configuration for the data source.
    """

    def __init__(
        self,
        query: str | None = None,
        database: str | None = None,
        group: str | None = None,
        table: str | None = None,
        path: str | None = None,
        storage_connector: sc.StorageConnector | dict[str, Any] | None = None,
        metrics: list[str] | None = None,
        dimensions: list[str] | None = None,
        rest_endpoint: RestEndpointConfig | dict | None = None,
        **kwargs,
    ):
        """Initialize a DataSource object.

        Parameters:
            query: SQL query string for the data source, if applicable.
            database: Name of the database containing the data source.
            group: Group or schema name for the data source.
            table: Table name for the data source.
            path: File system path for the data source.
            storage_connector: Storage connector object holds configuration for accessing the data source.
            metrics: List of metric column names for the data source.
            dimensions: List of dimension column names for the data source.
            rest_endpoint: REST endpoint configuration for the data source.
            **kwargs: Additional keyword arguments.
        """
        self._query = query
        self._database = database
        self._group = group
        self._table = table
        self._path = path
        if storage_connector is not None and isinstance(storage_connector, dict):
            self._storage_connector = sc.StorageConnector.from_response_json(
                storage_connector
            )
        else:
            self._storage_connector: sc.StorageConnector = storage_connector
        self._metrics = metrics or []
        self._dimensions = dimensions or []
        self._rest_endpoint = (
            RestEndpointConfig.from_response_json(rest_endpoint)
            if isinstance(rest_endpoint, dict)
            else rest_endpoint
        )

    @classmethod
    def from_response_json(
        cls,
        json_dict: dict[str, Any],
        storage_connector: sc.StorageConnector | None = None,
    ) -> DataSource | list[DataSource] | None:
        """Create a DataSource object (or list of objects) from a JSON response.

        Parameters:
            json_dict: The JSON dictionary from the API response.
            storage_connector: The storage connector object.

        Returns:
            The created object(s), or None if input is None.
        """
        if json_dict is None:
            return None  # TODO: change to [] and fix the tests

        json_decamelized: dict = humps.decamelize(json_dict)

        if "items" not in json_decamelized:
            data_source = cls(**json_decamelized)
            if storage_connector is not None:
                data_source.storage_connector = storage_connector
            return data_source

        return [
            DataSource.from_response_json(item, storage_connector)
            for item in json_decamelized["items"]
        ]

    def to_dict(self) -> dict:
        """Convert the DataSource object to a dictionary.

        Returns:
            Dictionary representation of the object.
        """
        ds_meta_dict = {
            "query": self._query,
            "database": self._database,
            "group": self._group,
            "table": self._table,
            "path": self._path,
            "metrics": self._metrics,
            "dimensions": self._dimensions,
            "restEndpoint": (
                self._rest_endpoint.to_dict() if self._rest_endpoint else None
            ),
        }
        if self._storage_connector:
            ds_meta_dict["storageConnector"] = self._storage_connector.to_dict()
        return ds_meta_dict

    def json(self) -> str:
        """Serialize the DataSource object to a JSON string.

        Returns:
            JSON string representation of the object.
        """
        return json.dumps(self, cls=util.Encoder)

    @public
    @property
    def query(self) -> str | None:
        """Get or set the SQL query string for the data source."""
        return self._query

    @query.setter
    def query(self, query: str) -> None:
        self._query = query

    @public
    @property
    def database(self) -> str | None:
        """Get or set the database name for the data source."""
        return self._database

    @database.setter
    def database(self, database: str) -> None:
        self._database = database

    @public
    @property
    def group(self) -> str | None:
        """Get or set the group/schema name for the data source."""
        return self._group

    @group.setter
    def group(self, group: str) -> None:
        self._group = group

    @public
    @property
    def table(self) -> str | None:
        """Get or set the table name for the data source."""
        return self._table

    @table.setter
    def table(self, table: str) -> None:
        self._table = table

    @public
    @property
    def path(self) -> str | None:
        """Get or set the file system path for the data source."""
        return self._path

    @path.setter
    def path(self, path: str) -> None:
        self._path = path

    @public
    @property
    def storage_connector(self) -> sc.StorageConnector | None:
        """Get or set the storage connector for the data source."""
        return self._storage_connector

    @storage_connector.setter
    def storage_connector(self, storage_connector: sc.StorageConnector) -> None:
        self._storage_connector = storage_connector

    @public
    def get_databases(self) -> list[str]:
        """Retrieve the list of available databases.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            data_source = fs.get_data_source("test_data_source")

            databases = data_source.get_databases()
            ```

        Returns:
            A list of database names available in the data source.
        """
        return self._storage_connector.get_databases()

    @public
    def get_tables(self, database: str | None = None) -> list[DataSource]:
        """Retrieve the list of tables from the specified database.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            data_source = fs.get_data_source("test_data_source")

            tables = data_source.get_tables()
            ```

        Parameters:
            database:
                The name of the database to list tables from.
                If not provided, the default database is used.

        Returns:
            A list of DataSource objects representing the tables.
        """
        return self._storage_connector.get_tables(database)

    @public
    def get_data(self, use_cached: bool = True) -> dsd.DataSourceData:
        """Retrieve the data from the data source.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            table = fs.get_data_source("test_data_source").get_tables()[0]

            data = table.get_data()
            ```

        Parameters:
            use_cached:
                Whether to use cached data if available.
                Only supported for CRM and REST connectors.
                Defaults to `True`.

        Returns:
            An object containing the data retrieved from the data source.
        """
        return self._storage_connector.get_data(self, use_cached=use_cached)

    @public
    def get_metadata(self) -> dict:
        """Retrieve metadata information about the data source.

        Example:
            ```python
            # connect to the Feature Store
            fs = ...

            table = fs.get_data_source("test_data_source").get_tables()[0]

            metadata = table.get_metadata()
            ```

        Returns:
            A dictionary containing metadata about the data source.
        """
        return self._storage_connector.get_metadata(self)

    @public
    def get_feature_groups_provenance(self) -> Links | None:
        """Get the generated feature groups using this data source, based on explicit provenance.

        These feature groups can be accessible or inaccessible.
        Explicit provenance does not track deleted generated feature group links, so deleted will always be empty.
        For inaccessible feature groups, only a minimal information is returned.

        Returns:
            The feature groups generated using this data source or `None` if none were created.

        Raises:
            hopsworks.client.exceptions.RestAPIError: In case the backend encounters an issue.
        """
        return self._storage_connector.get_feature_groups_provenance()

    @public
    def get_feature_groups(self) -> list[fg.FeatureGroup]:
        """Get the feature groups using this data source, based on explicit provenance.

        Only the accessible feature groups are returned.
        For more items use the base method, [`DataSource.get_feature_groups_provenance`][hsfs.core.data_source.DataSource.get_feature_groups_provenance].

        Returns:
            List of feature groups.
        """
        return self._storage_connector.get_feature_groups()

    @public
    def get_training_datasets_provenance(self) -> Links:
        """Get the generated training datasets using this data source, based on explicit provenance.

        These training datasets can be accessible or inaccessible.
        Explicit provenance does not track deleted generated training dataset links, so deleted will always be empty.
        For inaccessible training datasets, only a minimal information is returned.

        Returns:
            The training datasets generated using this data source or `None` if none were created.

        Raises:
            hopsworks.client.exceptions.RestAPIError: In case the backend encounters an issue.
        """
        return self._storage_connector.get_training_datasets_provenance()

    @public
    def get_training_datasets(self) -> list[TrainingDataset]:
        """Get the training datasets using this data source, based on explicit provenance.

        Only the accessible training datasets are returned.
        For more items use the base method, [`get_training_datasets_provenance`][hsfs.core.data_source.DataSource.get_training_datasets_provenance].

        Returns:
            List of training datasets.
        """
        return self._storage_connector.get_training_datasets()

    @public
    @property
    def metrics(self) -> list[str]:
        return self._metrics

    @metrics.setter
    def metrics(self, metrics: list[str]) -> None:
        self._metrics = metrics

    @public
    @property
    def dimensions(self) -> list[str]:
        return self._dimensions

    @dimensions.setter
    def dimensions(self, dimensions: list[str]) -> None:
        self._dimensions = dimensions

    @public
    @property
    def rest_endpoint(self) -> RestEndpointConfig | None:
        return self._rest_endpoint

    @rest_endpoint.setter
    def rest_endpoint(self, rest_endpoint: RestEndpointConfig) -> None:
        self._rest_endpoint = rest_endpoint

    def _update_storage_connector(self, storage_connector: sc.StorageConnector):
        """Update the storage connector configuration using DataSource.

        This internal method updates the connectors target database, schema,
        and table to match the information stored in the provided DataSource object.

        Parameters:
            storage_connector: A StorageConnector instance to be updated depending on the connector type.
        """
        if not storage_connector:
            return

        if storage_connector.type == sc.StorageConnector.REDSHIFT:
            if self.database:
                storage_connector._database_name = self.database
            if self.group:
                storage_connector._database_group = self.group
            if self.table:
                storage_connector._table_name = self.table
        if storage_connector.type == sc.StorageConnector.SNOWFLAKE:
            if self.database:
                storage_connector._database = self.database
            if self.group:
                storage_connector._schema = self.group
            if self.table:
                storage_connector._table = self.table
        if storage_connector.type == sc.StorageConnector.BIGQUERY:
            if self.database:
                storage_connector._query_project = self.database
            if self.group:
                storage_connector._dataset = self.group
            if self.table:
                storage_connector._query_table = self.table
        if storage_connector.type == sc.StorageConnector.SQL and self.database:
            storage_connector._database = self.database



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

import pandas as pd
from hopsworks_common.client import exceptions
from hsfs import (
    engine,
    feature_view,
    statistics,
    training_dataset,
    transformation_function,
)
from hsfs.core import transformation_function_api
from hsfs.hopsworks_udf import HopsworksUdf, UDFExecutionMode


if TYPE_CHECKING:
    import polars as pl
    import pyspark.sql as spark_sql
    from hsfs import feature_view, statistics, training_dataset, transformation_function


class TransformationFunctionEngine:
    BUILTIN_FN_NAMES = [
        "min_max_scaler",
        "standard_scaler",
        "robust_scaler",
        "label_encoder",
    ]
    AMBIGUOUS_FEATURE_ERROR = (
        "Provided feature '{}' in transformation functions is ambiguous and exists in more than one feature groups."
        "You can provide the feature with the prefix that was specified in the join."
    )
    FEATURE_NOT_EXIST_ERROR = "Provided feature '{}' in transformation functions do not exist in any of the feature groups."

    def __init__(self, feature_store_id: int):
        self._feature_store_id = feature_store_id
        self._transformation_function_api: transformation_function_api.TransformationFunctionApi = transformation_function_api.TransformationFunctionApi(
            feature_store_id
        )

    def save(
        self, transformation_fn_instance: transformation_function.TransformationFunction
    ):
        """Save a transformation function into the feature store.

        Parameters:
            transformation_fn_instance: The transformation function to be saved into the feature store.
        """
        self._transformation_function_api.register_transformation_fn(
            transformation_fn_instance
        )

    def get_transformation_fn(
        self, name: str, version: int | None = None
    ) -> (
        transformation_function.TransformationFunction
        | list[transformation_function.TransformationFunction]
    ):
        """Retrieve a transformation function from the feature store.

        If only the name of the transformation function is provided then all the versions of the transformation functions are returned as a list.
        If both name and version are not provided then all transformation functions saved in the feature view is returned.

        Parameters:
            name: The name of the transformation function to be retrieved.
            version: The version of the transformation function to be retrieved.

        Returns:
            A transformation function if name and version is provided.
            A list of transformation functions if only name is provided.
        """
        return self._transformation_function_api.get_transformation_fn(name, version)

    def get_transformation_fns(
        self,
    ) -> list[transformation_function.TransformationFunction]:
        """Get all the transformation functions in the feature store.

        Returns:
            A list of transformation functions.
        """
        transformation_fn_instances = (
            self._transformation_function_api.get_transformation_fn(
                name=None, version=None
            )
        )
        transformation_fns = []
        for (
            transformation_fn_instance
        ) in transformation_fn_instances:  # todo what is the point of this?
            transformation_fns.append(transformation_fn_instance)
        return transformation_fns

    @staticmethod
    def _validate_transformation_function_arguments(
        transformation_functions: list[transformation_function.TransformationFunction],
        data: spark_sql.DataFrame | pl.DataFrame | pd.DataFrame | dict[str, Any],
        request_parameters: dict[str, Any] | None = None,
    ) -> None:
        """Function to validate if all arguments required to execute the transformation functions are present are present in the passed data or request parameters.

        Parameters:
            transformation_functions: List of transformation functions to validate.
            data: The dataframe or list of dictionaries to validate the transformation functions against.
            request_parameters: Request parameters to validate the transformation functions against.

        Raises:
            exceptions.TransformationFunctionException: If the arguments required to execute the transformation functions are not present in the passed data or request parameters.
        """
        for tf in transformation_functions:
            if engine.get_instance().check_supported_dataframe(data):
                missing_features = set(tf.hopsworks_udf.transformation_features) - set(
                    data.columns
                )
            elif isinstance(data, dict):
                missing_features = set(tf.hopsworks_udf.transformation_features) - set(
                    data.keys()
                )
            else:
                raise exceptions.FeatureStoreException(
                    f"Dataframe type {type(data)} not supported in the engine."
                )

            if request_parameters:
                missing_features = missing_features - set(request_parameters.keys())
                if tf.hopsworks_udf.feature_name_prefix:
                    missing_features = missing_features - {
                        tf.hopsworks_udf.feature_name_prefix + feature
                        for feature in request_parameters
                    }

            if missing_features:
                raise exceptions.TransformationFunctionException(
                    message=f"The following feature(s): `{', '.join(missing_features)}`, required for the transformation function '{tf.hopsworks_udf.function_name}' are not available.",
                    missing_features=missing_features,
                    transformation_function_name=tf.hopsworks_udf.function_name,
                    transformation_type=tf.transformation_type.value,
                )

    @staticmethod
    def apply_transformation_functions(
        transformation_functions: list[transformation_function.TransformationFunction],
        data: spark_sql.DataFrame | pl.DataFrame | pd.DataFrame | dict[str, Any],
        online: bool = False,
        transformation_context: dict[str, Any] | list[dict[str, Any]] = None,
        request_parameters: dict[str, Any] = None,
        expected_features: set[str] = None,
    ) -> list[dict[str, Any]] | pd.DataFrame | pl.DataFrame:
        """Function to apply the transformation functions to the passed data.

        This function validates the arguments and calls the required function to apply the transformation functions to the passed data.
        For spark engine, the transformation functions are pushed down to Spark and is completely handled by the Spark engine.

        Parameters:
            transformation_functions: List of transformation functions to apply.
            data: The dataframe or list of dictionaries to apply the transformations to.
            online: Apply the transformations for online or offline usecase. This parameter is applicable when a transformation function is defined using the `default` execution mode.
            transformation_context: Transformation context to be used when applying the transformations.
            request_parameters: Request parameters to be used when applying the transformations.
            expected_features: Expected features to be present in the data, this is required to avoid dropping features with same names that are available from other feature groups in a feature view.

        Returns:
            The updated dataframe or list of dictionaries with the transformations applied.
        """
        if not transformation_functions or data is None:
            return data
        TransformationFunctionEngine._validate_transformation_function_arguments(
            transformation_functions=transformation_functions,
            data=data,
            request_parameters=request_parameters,
        )

        if isinstance(data, dict) or engine.get_type() != "spark":
            # If the data is a dictionary or if the engine is not spark, we execute the transformation functions using.
            return TransformationFunctionEngine._apply_transformation_functions(
                transformation_functions=transformation_functions,
                data=data,
                online=online,
                transformation_context=transformation_context,
                request_parameters=request_parameters,
                expected_features=expected_features,
            )
        # In the case of spark, we execute the transformation functions using the spark engine since the transformations are pushed down to Spark and are not executed in Python.
        return engine.get_instance()._apply_transformation_function(
            transformation_functions=transformation_functions,
            dataset=data,
            transformation_context=transformation_context,
            expected_features=expected_features,
        )

    @staticmethod
    def _apply_transformation_functions(
        transformation_functions: list[transformation_function.TransformationFunction],
        data: spark_sql.DataFrame | pl.DataFrame | pd.DataFrame | dict[str, Any],
        online: bool = False,
        transformation_context: dict[str, Any] | list[dict[str, Any]] = None,
        request_parameters: dict[str, Any] = None,
        expected_features: set[str] = None,
    ) -> list[dict[str, Any]] | pd.DataFrame | pl.DataFrame:
        """Function to apply the transformation functions to the passed dataframe or list of dictionaries.

        This function is only used when the engine is python or if the passed data is a dictionary.

        Parameters:
            transformation_functions: List of transformation functions to apply.
            data: The dataframe or list of dictionaries to apply the transformations to.
            online: Apply the transformations for online or offline usecase. This parameter is applicable when a transformation function is defined using the `default` execution mode.
            transformation_context: Transformation context to be used when applying the transformations.
            request_parameters: Request parameters to be used when applying the transformations.
            expected_features: Expected features to be present in the data, this is required to avoid dropping features with same names that are available from other feature groups in a feature view.

        Returns:
            The updated dataframe or list of dictionaries with the transformations applied.
        """
        dropped_features: set[str] = set()

        if isinstance(data, dict):
            transformed_data = data.copy()
        else:
            transformed_data = engine.get_instance().shallow_copy_dataframe(data)

        if request_parameters:
            for key in request_parameters:
                transformed_data[key] = request_parameters[key]

        for tf in transformation_functions:
            udf = tf.hopsworks_udf
            udf.transformation_context = transformation_context

            if udf.dropped_features:
                dropped_features.update(
                    {f for f in udf.dropped_features if f not in expected_features}
                    if expected_features
                    else udf.dropped_features
                )  # Drop features that are not expected, this is required to avoid dropping features having same name that are available from other feature groups.

            transformed_data = TransformationFunctionEngine.execute_udf(
                udf=udf, data=transformed_data, online=online
            )

        if isinstance(transformed_data, dict):
            transformed_data = {
                k: v for k, v in transformed_data.items() if k not in dropped_features
            }
        else:
            transformed_data = engine.get_instance().drop_columns(
                transformed_data, dropped_features
            )

        return transformed_data

    @staticmethod
    def execute_udf(
        udf: HopsworksUdf,
        data: spark_sql.DataFrame | pl.DataFrame | pd.DataFrame | dict[str, Any],
        online: bool = False,
    ) -> list[dict[str, Any]] | pd.DataFrame | pl.DataFrame:
        """Function to execute the UDF used to defined a transformation function on passed data.

        The functions pushes the execution of Pandas and Python Dataframes to the Python Engine and handles the execution dictionaries.

        Parameters:
            udf: The transformation function to execute.
            data: The dataframe or list of dictionaries to execute the transformation function on.
            online: Apply the transformations for online or offline usecase. This parameter is applicable when a transformation function is defined using the `default` execution mode.

        Returns:
            The updated dataframe or list of dictionaries with the transformations applied.
        """
        execution_engine = engine.get_instance()
        if execution_engine.check_supported_dataframe(data):
            return execution_engine.apply_udf_on_dataframe(
                udf=udf, dataframe=data, online=online
            )
        if isinstance(data, dict):
            return TransformationFunctionEngine.apply_udf_on_dict(
                udf=udf, data=data, online=online
            )
        raise exceptions.FeatureStoreException(
            f"Dataframe type {type(data)} not supported in the engine."
        )

    @staticmethod
    def apply_udf_on_dict(
        udf: HopsworksUdf,
        data: dict[str, Any],
        online: bool | None = True,
    ) -> dict[str, Any]:
        """Function to apply the UDF used to defined a transformation function on a dictionary.

        The function is not pushed to the Python Engine since it should be executed this function in both the Python and Spark Kernel.

        Parameters:
            udf: The transformation function to execute.
            data: The dictionary to execute the transformation function on.
            online: Apply the transformations for online or offline usecase. This parameter is applicable when a transformation function is defined using the `default` execution mode.

        Returns:
            The updated dictionary with the transformations applied.
        """
        features = []

        if not online and engine.get_type() == "spark":
            raise exceptions.FeatureStoreException(
                "Cannot apply transformation functions on a dictionary in offline mode when the engine is spark. Please use the python engine or use the online mode."
            )

        for unprefixed_feature in udf.unprefixed_transformation_features:
            prefixed_feature = (
                udf.feature_name_prefix + unprefixed_feature
                if udf.feature_name_prefix
                else unprefixed_feature
            )
            feature_value = data.get(prefixed_feature, data.get(unprefixed_feature))

            if (
                udf.execution_mode.get_current_execution_mode(online=online)
                == UDFExecutionMode.PANDAS
            ):
                features.append(pd.Series(feature_value))
            else:
                features.append(feature_value)

        transformed_result = udf.get_udf(online=online)(*features)

        if (
            udf.execution_mode.get_current_execution_mode(online=online)
            == UDFExecutionMode.PANDAS
        ):
            # Pandas UDF return can return a pandas series or a pandas dataframe, so we need to cast it back to a dictionary.
            if isinstance(transformed_result, pd.Series):
                data[transformed_result.name] = transformed_result.values[0]
            else:
                for col in transformed_result:
                    data[col] = transformed_result[col].values[0]
        else:
            # Python UDF return can return a tuple or a list, so we need to cast it back to a dictionary.
            if isinstance(transformed_result, (tuple, list)):
                for index, result in enumerate(transformed_result):
                    data[udf.output_column_names[index]] = result
            else:
                data[udf.output_column_names[0]] = transformed_result

        return data

    def delete(
        self,
        transformation_function_instance: transformation_function.TransformationFunction,
    ) -> None:
        """Delete a transformation function from the feature store.

        Parameters:
            transformation_function_instance: The transformation function to be removed from the feature store.
        """
        self._transformation_function_api.delete(transformation_function_instance)

    @staticmethod
    def compute_transformation_fn_statistics(
        training_dataset_obj: training_dataset.TrainingDataset,
        statistics_features: list[str],
        label_encoder_features: list[str],
        feature_dataframe: pd.DataFrame
        | pl.DataFrame
        | TypeVar("pyspark.sql.DataFrame"),
        feature_view_obj: feature_view.FeatureView,
    ) -> statistics.Statistics:
        """Compute the statistics required for a training dataset object.

        Parameters:
            training_dataset_obj : The training dataset for which the statistics is to be computed.
            statistics_features : The list of features for which the statistics should be computed.
            label_encoder_features : Features used for label encoding.
            feature_dataframe : The dataframe that contains the data for which the statistics must be computed.
            feature_view_obj : The feature view in which the training data is being created.

        Returns:
            The statistics object that contains the statistics for each features.
        """
        return training_dataset_obj._statistics_engine.compute_transformation_fn_statistics(
            td_metadata_instance=training_dataset_obj,
            columns=statistics_features,
            label_encoder_features=label_encoder_features,  # label encoded features only
            feature_dataframe=feature_dataframe,
            feature_view_obj=feature_view_obj,
        )

    @staticmethod
    def get_ready_to_use_transformation_fns(
        feature_view: feature_view.FeatureView,
        training_dataset_version: int | None = None,
    ) -> list[transformation_function.TransformationFunction]:
        """Function that updates statistics required for all transformation functions in the feature view based on training dataset version.

        Parameters:
            feature_view: The feature view in which the training data is being created.
            training_dataset_version: The training version used to update the statistics used in the transformation functions.

        Returns:
            List of transformation functions.
        """
        # check if transformation functions require statistics
        is_stat_required = any(
            tf.hopsworks_udf.statistics_required
            for tf in feature_view.transformation_functions
        )
        if not is_stat_required:
            td_tffn_stats = None
        else:
            # if there are any transformation functions that require statistics get related statistics and
            # populate with relevant arguments
            # there should be only one statistics object with before_transformation=true
            if training_dataset_version is None:
                raise ValueError(
                    "Training data version is required for transformation. Call `feature_view.init_serving(version)` "
                    "or `feature_view.init_batch_scoring(version)` to pass the training dataset version."
                    "Training data can be created by `feature_view.create_training_data` or `feature_view.training_data`."
                )
            td_tffn_stats = feature_view._statistics_engine.get(
                feature_view,
                before_transformation=True,
                training_dataset_version=training_dataset_version,
            )

        if is_stat_required and td_tffn_stats is None:
            raise ValueError(
                "No statistics available for initializing transformation functions."
                "Training data can be created by `feature_view.create_training_data` or `feature_view.training_data`."
            )

        if is_stat_required:
            for transformation_function in feature_view.transformation_functions:
                transformation_function.transformation_statistics = (
                    td_tffn_stats.feature_descriptive_statistics
                )
        return feature_view.transformation_functions

    @staticmethod
    def compute_and_set_feature_statistics(
        training_dataset: training_dataset.TrainingDataset,
        feature_view_obj: feature_view.FeatureView,
        dataset: dict[
            str, pd.DataFrame | pl.DataFrame | TypeVar("pyspark.sql.DataFrame")
        ]
        | pd.DataFrame
        | pl.DataFrame
        | TypeVar("pyspark.sql.DataFrame"),
    ) -> None:
        """Function that computes and sets the statistics required for the UDF used for transformation.

        The function assigns the statistics computed to hopsworks UDF object so that the statistics can be used when UDF is executed.

        Parameters:
            training_dataset : The training dataset for which the statistics is to be computed.
            feature_view_obj : The feature view in which the training data is being created.
            dataset : A dataframe that contains the training data or a dictionary that contains both the training and test data.
        """
        statistics_features: set[str] = set()
        label_encoder_features: set[str] = set()

        # Finding the features for which statistics is required
        for tf in feature_view_obj.transformation_functions:
            statistics_features.update(tf.hopsworks_udf.statistics_features)
            if (
                tf.hopsworks_udf.function_name == "label_encoder"
                or tf.hopsworks_udf.function_name == "one_hot_encoder"
            ):
                label_encoder_features.update(tf.hopsworks_udf.statistics_features)
        if statistics_features:
            # compute statistics on training data
            if training_dataset.splits:
                # compute statistics before transformations are applied
                stats = (
                    TransformationFunctionEngine.compute_transformation_fn_statistics(
                        training_dataset,
                        list(statistics_features),
                        list(label_encoder_features),
                        dataset.get(training_dataset.train_split),
                        feature_view_obj,
                    )
                )
            else:
                stats = (
                    TransformationFunctionEngine.compute_transformation_fn_statistics(
                        training_dataset,
                        list(statistics_features),
                        list(label_encoder_features),
                        dataset,
                        feature_view_obj,
                    )
                )

            # Set statistics computed in the hopsworks UDF
            for tf in feature_view_obj.transformation_functions:
                tf.transformation_statistics = stats.feature_descriptive_statistics

    @staticmethod
    def get_and_set_feature_statistics(
        training_dataset: training_dataset.TrainingDataset,
        feature_view_obj: feature_view.FeatureView,
        training_dataset_version: int = None,
    ) -> None:
        """Function that gets the transformation statistics computed while creating the training dataset from the backend and assigns it to the hopsworks UDF object.

        The function assigns the statistics computed to hopsworks UDF object so that the statistics can be used when UDF is executed.

        Parameters:
            training_dataset: The training dataset for which the statistics is to be computed.
            feature_view_obj: The feature view in which the training data is being created.
            training_dataset_version: The version of the training dataset for which the statistics is to be retrieved.

        Raises:
            ValueError: If the statistics are not present in the backend.
        """
        is_stat_required = any(
            tf.hopsworks_udf.statistics_required
            for tf in feature_view_obj.transformation_functions
        )

        if is_stat_required:
            td_tffn_stats = training_dataset._statistics_engine.get(
                feature_view_obj,
                before_transformation=True,
                training_dataset_version=training_dataset_version,
            )

            if td_tffn_stats is None:
                raise ValueError(
                    "No statistics available for initializing transformation functions."
                )

            for tf in feature_view_obj.transformation_functions:
                tf.transformation_statistics = (
                    td_tffn_stats.feature_descriptive_statistics
                )



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import json
from typing import TYPE_CHECKING

from hopsworks_apigen import also_available_as
from hopsworks_common import client, decorators, tag, usage


if TYPE_CHECKING:
    from hsfs.feature_group import FeatureGroup
    from hsfs.training_dataset import TrainingDataset


@also_available_as("hopsworks.core.tags_api.TagsApi", "hsfs.core.tags_api.TagsApi")
class TagsApi:
    def __init__(self, feature_store_id: int, entity_type: str):
        """Tags endpoint for `trainingdatasets` and `featuregroups` resource.

        Parameters:
            feature_store_id: id of the respective featurestore
            entity_type: "trainingdatasets" or "featuregroups"
        """
        self._feature_store_id = feature_store_id
        self._entity_type = entity_type

    @usage.method_logger
    def add(
        self,
        metadata_instance: TrainingDataset | FeatureGroup,
        name: str,
        value: str,
        training_dataset_version=None,
    ):
        """Attach a name/value tag to a training dataset or feature group.

        A tag consists of a name/value pair. Tag names are unique identifiers.
        The value of a tag can be any valid json - primitives, arrays or json objects.

        Parameters:
            metadata_instance: Metadata object of the instance to add the tag for.
            name: Name of the tag to be added.
            value: Value of the tag to be added.
            training_dataset_version: Version of the training dataset.
        """
        _client = client.get_instance()
        path_params = self.get_path(metadata_instance, training_dataset_version) + [
            name
        ]
        headers = {"content-type": "application/json"}
        json_value = json.dumps(value)
        _client._send_request("PUT", path_params, headers=headers, data=json_value)

    @usage.method_logger
    def delete(
        self,
        metadata_instance: TrainingDataset | FeatureGroup,
        name: str,
        training_dataset_version=None,
    ):
        """Delete a tag from a training dataset or feature group.

        Tag names are unique identifiers.

        Parameters:
            metadata_instance: Metadata object of training dataset to delete the tag for.
            name: Name of the tag to be removed.
            training_dataset_version: Version of the training dataset.
        """
        _client = client.get_instance()
        path_params = self.get_path(metadata_instance, training_dataset_version) + [
            name
        ]

        _client._send_request("DELETE", path_params)

    @usage.method_logger
    @decorators.catch_not_found("hopsworks_common.tag.Tag", fallback_return={})
    def get(
        self,
        metadata_instance: TrainingDataset | FeatureGroup,
        name: str | None = None,
        training_dataset_version=None,
    ) -> dict:
        """Get the tags of a training dataset or feature group.

        Gets all tags if no tag name is specified.

        Parameters:
            metadata_instance: Metadata object of training dataset to get the tags for.
            name: Tag name.
            training_dataset_version: Version of the training dataset.

        Returns:
            Dict of tag name/values.
        """
        _client = client.get_instance()
        path_params = self.get_path(metadata_instance, training_dataset_version)

        if name is not None:
            path_params.append(name)

        return {
            tag._name: json.loads(tag._value)
            for tag in tag.Tag.from_response_json(
                _client._send_request("GET", path_params)
            )
        }

    @usage.method_logger
    def get_path(self, metadata_instance, training_dataset_version=None):
        _client = client.get_instance()
        if hasattr(metadata_instance, "training_data"):
            # Only FeatureView has training_data method
            path = [
                "project",
                _client._project_id,
                "featurestores",
                self._feature_store_id,
                "featureview",
                metadata_instance.name,
                "version",
                metadata_instance.version,
            ]
            if training_dataset_version:
                return path + [
                    "trainingdatasets",
                    "version",
                    training_dataset_version,
                    "tags",
                ]
            return path + ["tags"]
        return [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            self._entity_type,
            metadata_instance.id,
            "tags",
        ]



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import grpc
from hopsworks_apigen import also_available_as
from hopsworks_common.client.istio.grpc.proto.grpc_predict_v2_pb2_grpc import (
    GRPCInferenceServiceStub,
)
from hopsworks_common.client.istio.utils.infer_type import InferRequest, InferResponse


@also_available_as("hsml.client.istio.grpc.inference_client._PathPrefixInterceptor")
class _PathPrefixInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Interceptor that prepends a path prefix to gRPC method calls for path-based routing."""

    def __init__(self, path_prefix: str):
        self._path_prefix = path_prefix

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_method = self._path_prefix + client_call_details.method
        new_details = _ClientCallDetails(
            method=new_method,
            timeout=client_call_details.timeout,
            metadata=client_call_details.metadata,
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
            compression=client_call_details.compression,
        )
        return continuation(new_details, request)


@also_available_as("hsml.client.istio.grpc.inference_client._ClientCallDetails")
class _ClientCallDetails(
    grpc.ClientCallDetails,
):
    """Implementation of grpc.ClientCallDetails for use by interceptors."""

    def __init__(
        self, method, timeout, metadata, credentials, wait_for_ready, compression
    ):
        self.method = method
        self.timeout = timeout
        self.metadata = metadata
        self.credentials = credentials
        self.wait_for_ready = wait_for_ready
        self.compression = compression


@also_available_as("hsml.client.istio.grpc.inference_client.GRPCInferenceServerClient")
class GRPCInferenceServerClient:
    def __init__(
        self,
        url,
        serving_api_key,
        path_prefix=None,
    ):
        channel_opt = [
            ("grpc.max_send_message_length", -1),
            ("grpc.max_receive_message_length", -1),
        ]

        # Authentication is done via API Key in the Authorization header
        self._channel = grpc.insecure_channel(url, options=channel_opt)

        # Apply path prefix interceptor for path-based routing
        if path_prefix:
            self._channel = grpc.intercept_channel(
                self._channel, _PathPrefixInterceptor(path_prefix)
            )

        self._client_stub = GRPCInferenceServiceStub(self._channel)
        self._serving_api_key = serving_api_key

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __del__(self):
        """It is called during object garbage collection."""
        self.close()

    def close(self):
        """Close the client. Future calls to server will result in an Error."""
        self._channel.close()

    def infer(self, infer_request: InferRequest, headers=None, client_timeout=None):
        headers = {} if headers is None else headers
        headers["authorization"] = "ApiKey " + self._serving_api_key
        metadata = headers.items()

        # convert the InferRequest to a ModelInferRequest message
        request = infer_request.to_grpc()

        try:
            # send request
            model_infer_response = self._client_stub.ModelInfer(
                request=request, metadata=metadata, timeout=client_timeout
            )
        except grpc.RpcError as rpc_error:
            raise rpc_error

        # convert back the ModelInferResponse message to InferResponse
        return InferResponse.from_grpc(model_infer_response)



#
#   Copyright 2023 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import contextlib
import logging
import re
import threading
from functools import wraps

import opensearchpy
import urllib3
from hopsworks_apigen import also_available_as, public
from hopsworks_common.client.exceptions import (
    FeatureStoreException,
    VectorDatabaseException,
)
from hopsworks_common.core.opensearch_api import OPENSEARCH_CONFIG, OpenSearchApi
from opensearchpy import OpenSearch
from opensearchpy.exceptions import (
    AuthenticationException,
    ConnectionError,
    ConnectionTimeout,
    RequestError,
)
from retrying import retry


def _is_timeout(exception):
    return isinstance(
        exception, (urllib3.exceptions.ReadTimeoutError, ConnectionTimeout)
    )


def _handle_opensearch_exception(func):
    @wraps(func)
    def error_handler_wrapper(*args, **kw):
        try:
            return func(*args, **kw)
        except (ConnectionError, AuthenticationException):
            # OpenSearchConnectionError occurs when connection is closed.
            # OpenSearchAuthenticationException occurs when jwt is expired
            # args[0] is 'self' - the ProjectOpenSearchClient instance
            client_wrapper = args[0] if args else None
            if client_wrapper and isinstance(client_wrapper, ProjectOpenSearchClient):
                client_wrapper.refresh_opensearch_connection()
            return func(*args, **kw)
        except RequestError as e:
            caused_by = e.info.get("error") and e.info["error"].get("caused_by")
            if caused_by and caused_by["type"] == "illegal_argument_exception":
                client_wrapper = args[0] if args else None
                if client_wrapper and isinstance(
                    client_wrapper, ProjectOpenSearchClient
                ):
                    raise client_wrapper._create_vector_database_exception(
                        caused_by["reason"]
                    ) from e
                raise VectorDatabaseException(
                    VectorDatabaseException.OTHERS,
                    f"Error in Opensearch request: {caused_by['reason']}",
                    e.info,
                ) from e
            raise VectorDatabaseException(
                VectorDatabaseException.OTHERS,
                f"Error in Opensearch request: {e}",
                e.info,
            ) from e
        except Exception as e:
            if _is_timeout(e):
                raise FeatureStoreException(
                    ProjectOpenSearchClient.TIMEOUT_ERROR_MSG
                ) from e
            client_wrapper = args[0] if args else None
            # Invalidate the cache for federated connector clients if the error is not a timeout,
            # so that it wont't stuck at a error state.
            if (
                client_wrapper
                and isinstance(client_wrapper, ProjectOpenSearchClient)
                and not client_wrapper.is_cluster_client
            ):
                OpenSearchClientSingleton.invalidate_cache(
                    client_wrapper.feature_store_id, close_opensearch_client=True
                )
            raise e

    return error_handler_wrapper


@public(
    "hopsworks.core.opensearch.OpensearchRequestOption",
    "hsfs.core.opensearch.OpensearchRequestOption",
)
class OpensearchRequestOption:
    DEFAULT_OPTION_MAP = {
        "timeout": "30s",
    }

    DEFAULT_OPTION_MAP_V2_3 = {
        # starting from opensearch client v2.3, timeout should be in int/float
        # https://github.com/opensearch-project/opensearch-py/pull/394
        "timeout": 30,
    }

    @public
    @classmethod
    def get_version(cls) -> tuple[int, int]:
        """Get the major and minor version of the opensearch client.

        Returns:
            A tuple containing the major and minor version of the opensearch client.
        """
        return opensearchpy.__version__[0:2]

    @public
    @classmethod
    def get_options(cls, options: dict) -> dict:
        """Construct a map of options for the request to the vector database.

        Parameters:
            options:
                The options used for the request to the vector database.
                The keys are attribute values of the OpensearchRequestOption class.

        Returns:
            A dictionary containing the constructed options map, where keys represent attribute values of the OpensearchRequestOption class, and values are obtained either from the provided options or default values if not available.
        """
        default_option = (
            cls.DEFAULT_OPTION_MAP
            if cls.get_version() < (2, 3)
            else cls.DEFAULT_OPTION_MAP_V2_3
        )
        if options:
            # make lower case to avoid issues with cases
            options = {k.lower(): v for k, v in options.items()}
            new_options = {}
            for option, value in default_option.items():
                if option in options:
                    if (
                        option == "timeout"
                        and cls.get_version() < (2, 3)
                        and isinstance(options[option], int)
                    ):
                        new_options[option] = f"{options[option]}s"
                    else:
                        new_options[option] = options[option]
                else:
                    new_options[option] = value
            return new_options
        return default_option


class ProjectOpenSearchClient:
    """Wrapper for OpenSearch client associated with a specific project.

    Thread-safe and can be used concurrently.
    """

    TIMEOUT_ERROR_MSG = """
    Cannot fetch results from Opensearch due to timeout. It is because the server is busy right now or longer time is needed to reload a large index. Try and increase the timeout limit by providing the parameter `options={"timeout": 60}` in the method `find_neighbor` or `count`.
    """

    def __init__(
        self,
        opensearch_client: OpenSearch,
        feature_store_id: int = None,
        is_cluster_client: bool = True,
    ):
        self._opensearch_client = opensearch_client
        self._feature_store_id = feature_store_id
        self._is_cluster_client = is_cluster_client
        self._client_lock = threading.RLock()

    @retry(
        wait_exponential_multiplier=1000,
        stop_max_attempt_number=5,
        retry_on_exception=_is_timeout,
    )
    @_handle_opensearch_exception
    def search(self, index=None, body=None, options=None):
        return self.get_opensearch_client().search(
            body=body, index=index, params=OpensearchRequestOption.get_options(options)
        )

    @retry(
        wait_exponential_multiplier=1000,
        stop_max_attempt_number=5,
        retry_on_exception=_is_timeout,
    )
    @_handle_opensearch_exception
    def count(self, index, body=None, options=None):
        result = self.get_opensearch_client().count(
            index=index, body=body, params=OpensearchRequestOption.get_options(options)
        )
        return result["count"]

    def refresh_opensearch_connection(self):
        """Refresh the OpenSearch connection for the client."""
        if self.is_cluster_client:
            OpenSearchClientSingleton.get_instance().close_cluster_client()
        else:
            self.get_opensearch_client().close()
        self._opensearch_client = None
        # Recreate the client
        self.get_opensearch_client()

    def close(self):
        """Close the underlying OpenSearch client."""
        # Close the client if it is a federated connector client otherwise the cluster client is closed when python client is closed
        OpenSearchClientSingleton.invalidate_cache(
            self._feature_store_id, close_opensearch_client=self.is_cluster_client
        )

        # For default client, close the cluster client
        if not self._feature_store_id:
            OpenSearchClientSingleton.get_instance().close_cluster_client()
        self._opensearch_client = None

    def _create_vector_database_exception(self, message):
        """Create appropriate VectorDatabaseException based on error message."""
        if "[knn] requires k" in message:
            pattern = r"\[knn\] requires k <= (\d+)"
            match = re.search(pattern, message)
            if match:
                k = match.group(1)
                reason = VectorDatabaseException.REQUESTED_K_TOO_LARGE
                message = (
                    f"Illegal argument in vector database request: "
                    f"Requested k is too large, it needs to be less than {k}."
                )
                info = {VectorDatabaseException.REQUESTED_K_TOO_LARGE_INFO_K: int(k)}
            else:
                reason = VectorDatabaseException.REQUESTED_K_TOO_LARGE
                message = "Illegal argument in vector database request: Requested k is too large."
                info = {}
        elif "Result window is too large" in message:
            pattern = r"or equal to: \[(\d+)\]"
            match = re.search(pattern, message)
            if match:
                n = match.group(1)
                reason = VectorDatabaseException.REQUESTED_NUM_RESULT_TOO_LARGE
                message = (
                    f"Illegal argument in vector database request: "
                    f"Requested n is too large, it needs to be less than {n}."
                )
                info = {
                    VectorDatabaseException.REQUESTED_NUM_RESULT_TOO_LARGE_INFO_N: int(
                        n
                    )
                }
            else:
                reason = VectorDatabaseException.REQUESTED_NUM_RESULT_TOO_LARGE
                message = (
                    "Illegal argument in vector database request: "
                    "Requested n is too large."
                )
                info = {}
        else:
            reason = VectorDatabaseException.OTHERS
            message = message
            info = {}
        return VectorDatabaseException(reason, message, info)

    def get_opensearch_client(self):
        """Get the underlying OpenSearch client."""
        if self.is_cluster_client:
            return (
                OpenSearchClientSingleton.get_instance().get_or_create_cluster_client()
            )
        if (
            self._opensearch_client is None
        ):  # when refreshing the connection, opensearch_client is None
            with self._client_lock:
                if self._opensearch_client is None:
                    _, self._opensearch_client = (
                        OpenSearchClientSingleton.get_instance()._get_or_create_opensearch_client(
                            self._feature_store_id
                        )
                    )
        return self._opensearch_client

    @property
    def feature_store_id(self):
        return self._feature_store_id

    @property
    def is_cluster_client(self):
        return self._is_cluster_client


@also_available_as(
    "hopsworks.core.opensearch.OpenSearchClientSingleton",
    "hsfs.core.opensearch.OpenSearchClientSingleton",
)
class OpenSearchClientSingleton:
    """Thread-safe singleton for OpenSearch client reuse.

    Caches:
      - ProjectOpenSearchClient wrappers keyed by feature_store_id (or "default" if no feature store id is provided)
      - Federated connector configs per feature store
      - A single default/cluster OpenSearch client

    Client selection:
      - If a feature store has a federated OpenSearch connector, a dedicated OpenSearch
        client is created with its config and cached under that feature_store_id.
      - If no connector exists (or feature_store_id is None), the shared cluster
        OpenSearch client is used
    """

    _instance = None
    _default_opensearch_client = None
    _wrapper_cache = {}  # Cache ProjectOpenSearchClient wrappers by feature_store_id
    _cache_lock = threading.RLock()  # Reentrant lock for thread-safe cache access
    _federated_connector_cache = {}  # Cache federated connector check results
    _opensearch_api = None

    TIMEOUT_ERROR_MSG = ProjectOpenSearchClient.TIMEOUT_ERROR_MSG
    FEDERATED_CONNECTOR_NAME = "opensearch_connector"
    DEFAULT_CLUSTER_CACHE_KEY = "default"

    def __new__(cls, feature_store_id: int = None):
        if not cls._instance:
            with cls._cache_lock if hasattr(cls, "_cache_lock") else threading.RLock():
                if not cls._instance:  # Double-check locking
                    cls._instance = super().__new__(cls)
                    cls._instance._cache_lock = threading.RLock()
                    cls._instance._default_opensearch_client = None
                    cls._instance._federated_connector_cache = {}
                    cls._instance._opensearch_api = OpenSearchApi()

        # Return a ProjectOpenSearchClient wrapper for the requested feature_store_id
        return cls._instance._get_client_wrapper(feature_store_id)

    def _get_client_wrapper(self, feature_store_id: int = None):
        """Get or create a ProjectOpenSearchClient wrapper for the given feature_store_id."""
        cache_key = (
            f"fs_{feature_store_id}"
            if feature_store_id is not None
            else self.DEFAULT_CLUSTER_CACHE_KEY
        )
        if cache_key in self._wrapper_cache:
            return self._wrapper_cache[cache_key]
        with self._cache_lock:
            if cache_key not in self._wrapper_cache:
                is_cluster_client, opensearch_client = (
                    self._get_or_create_opensearch_client(feature_store_id)
                )
                wrapper = ProjectOpenSearchClient(
                    opensearch_client,
                    feature_store_id,
                    is_cluster_client=is_cluster_client,
                )
                self._wrapper_cache[cache_key] = wrapper
        return self._wrapper_cache[cache_key]

    def _get_federated_opensearch_config(self, feature_store_id: int) -> dict | None:
        """Try to fetch the federated_opensearch storage connector and return its config.

        Returns None if connector doesn't exist or isn't an OpenSearch connector.
        """
        cache_key = f"fs_{feature_store_id}" if feature_store_id else None

        if cache_key and cache_key in self._federated_connector_cache:
            return self._federated_connector_cache[cache_key]

        # Import here to avoid circular imports
        from hsfs.core import storage_connector_api

        connector_api = storage_connector_api.StorageConnectorApi()
        connector = connector_api.get(feature_store_id, self.FEDERATED_CONNECTOR_NAME)

        if connector is None:
            # Connector doesn't exist, do not cache anything
            return self._opensearch_api.get_default_py_config()

        # Check if it's an OpenSearch connector
        from hsfs.storage_connector import OpenSearchConnector

        if not isinstance(connector, OpenSearchConnector):
            logging.debug(
                f"Storage connector '{self.FEDERATED_CONNECTOR_NAME}' exists but is not an OpenSearch connector. "
                f"Using default OpenSearch configuration."
            )
            return None

        config = connector.connector_options()

        # Cache the config
        if cache_key:
            self._federated_connector_cache[cache_key] = config

        logging.debug(
            f"Using federated OpenSearch connector '{self.FEDERATED_CONNECTOR_NAME}' "
            f"with host: {connector.host}"
        )
        return config

    def _get_or_create_opensearch_client(self, feature_store_id: int = None):
        """Get cached client or create new one.

        If no federated connector config is found for a given feature store or no feature store id is provided,
        the client for that feature store reuses the default cluster client.
        """
        # query log is at INFO level
        # 2023-11-24 15:10:49,470 INFO: POST https://localhost:9200/index/_search [status:200 request:0.041s]
        logging.getLogger("opensearchpy").setLevel(logging.WARNING)
        logging.getLogger("opensearch").setLevel(logging.WARNING)

        if feature_store_id is None:
            return True, self.get_or_create_cluster_client()

        # Try to get federated connector config first
        opensearch_config = self._get_federated_opensearch_config(feature_store_id)

        if opensearch_config is None:
            return True, self.get_or_create_cluster_client()

        opensearch_config[OPENSEARCH_CONFIG.HEADERS] = {
            "Authorization": self._opensearch_api._get_authorization_token(
                feature_store_id
            )
        }

        # Dedicated client for this feature store
        return False, OpenSearch(**opensearch_config)

    def get_or_create_cluster_client(self):
        """Get or create a cluster client."""
        if self._default_opensearch_client is not None:
            return self._default_opensearch_client

        self._default_opensearch_client = OpenSearch(
            **OpenSearchApi().get_default_py_config()
        )
        return self._default_opensearch_client

    def close_cluster_client(self):
        """Close the cluster client."""
        with self._cache_lock:
            if self._default_opensearch_client is not None:
                self._default_opensearch_client.close()
                self._default_opensearch_client = None

    @classmethod
    def close_all(cls):
        """Close all cached OpenSearch clients. Thread-safe."""
        if cls._instance and hasattr(cls._instance, "_wrapper_cache"):
            with cls._instance._cache_lock:
                for client in list(cls._instance._wrapper_cache.values()):
                    with contextlib.suppress(Exception):
                        client.close()
                cls._instance._wrapper_cache.clear()
                # Also clear federated connector cache
                if hasattr(cls._instance, "_federated_connector_cache"):
                    cls._instance._federated_connector_cache.clear()
        with contextlib.suppress(Exception):
            cls._instance._default_opensearch_client.close()
            cls._instance._default_opensearch_client = None
        cls._instance = None

    @classmethod
    def get_instance(cls) -> OpenSearchClientSingleton:
        """Get the singleton instance.

        Returns:
            The singleton instance of OpenSearchClientSingleton.
        """
        return cls._instance

    @classmethod
    def invalidate_cache(
        cls, feature_store_id: int = None, close_opensearch_client: bool = False
    ):
        """Invalidate cached ProjectOpenSearchClient and connector config.

        Parameters:
            feature_store_id: The feature store ID to invalidate cache for.
            close_opensearch_client: Whether to close the cached OpenSearch client when invalidating.
        """
        if cls._instance is None:
            return

        if feature_store_id is not None:
            # Invalidate specific feature store cache
            fs_cache_key = f"fs_{feature_store_id}"
            closed_client = None
            with cls._instance._cache_lock:
                # remove the client from the cache
                if fs_cache_key in cls._instance._wrapper_cache:
                    if close_opensearch_client:
                        with contextlib.suppress(Exception):
                            closed_client = cls._instance._wrapper_cache[
                                fs_cache_key
                            ].get_opensearch_client()
                            closed_client.close()
                    del cls._instance._wrapper_cache[fs_cache_key]

                # Remove connector config cache
                if fs_cache_key in cls._instance._federated_connector_cache:
                    del cls._instance._federated_connector_cache[fs_cache_key]

                logging.debug(
                    f"Invalidated OpenSearch cache for feature store {feature_store_id}"
                )

                # If the closed client is the shared default client, clear it so a new one is created next time.
                if close_opensearch_client and closed_client is not None:
                    with contextlib.suppress(Exception):
                        if cls._instance._default_opensearch_client is closed_client:
                            cls._instance._default_opensearch_client = None



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from __future__ import annotations

import json

import humps
from hopsworks_apigen import public
from hopsworks_common import client, util
from hopsworks_common.constants import (
    INFERENCE_ENDPOINTS,
    MODEL,
    MODEL_SERVING,
    PREDICTOR,
    SCALING_CONFIG,
    Default,
)
from hsml import deployment
from hsml.deployable_component import DeployableComponent
from hsml.inference_batcher import InferenceBatcher
from hsml.inference_logger import InferenceLogger
from hsml.predictor_state import PredictorState
from hsml.resources import PredictorResources
from hsml.scaling_config import (
    PredictorScalingConfig,
)
from hsml.transformer import Transformer


@public
class Predictor(DeployableComponent):
    """Metadata object representing a predictor in Model Serving."""

    @staticmethod
    def _get_raw_num_instances(resources):
        if resources is None:
            return None
        return (
            resources._num_instances
            if hasattr(resources, "_num_instances")
            else resources.num_instances
        )

    def __init__(
        self,
        name: str,
        model_server: str,
        model_name: str = None,
        model_path: str = None,
        model_version: int = None,
        model_framework: str = None,  # MODEL.FRAMEWORK
        serving_tool: str | None = None,
        script_file: str | None = None,
        config_file: str | None = None,
        resources: PredictorResources | dict | Default | None = None,  # base
        inference_logger: InferenceLogger | dict | Default | None = None,  # base
        inference_batcher: InferenceBatcher | dict | Default | None = None,  # base
        transformer: Transformer | dict | Default | None = None,
        id: int | None = None,
        version: int | None = None,
        description: str | None = None,
        created_at: str | None = None,
        creator: str | None = None,
        api_protocol: str | None = INFERENCE_ENDPOINTS.API_PROTOCOL_REST,
        environment: str | None = None,
        project_namespace: str = None,
        scaling_configuration: PredictorScalingConfig | dict | Default | None = None,
        **kwargs,
    ):
        serving_tool = (
            self._validate_serving_tool(serving_tool)
            or self._get_default_serving_tool()
        )
        resources = self._validate_resources(
            util.get_obj_from_json(resources, PredictorResources), serving_tool
        ) or self._get_default_resources(serving_tool)

        self._scaling_configuration = util.get_obj_from_json(
            scaling_configuration, PredictorScalingConfig
        ) or PredictorScalingConfig.get_default_scaling_configuration(
            serving_tool=serving_tool,
            min_instances=self._get_raw_num_instances(resources),
        )

        super().__init__(
            script_file,
            resources,
            inference_batcher,
            scaling_configuration=self._scaling_configuration,
        )

        self._name = name
        self._model_name = model_name
        self._model_path = model_path
        self._model_version = model_version
        self._model_framework = model_framework
        self._serving_tool = serving_tool
        self._model_server = model_server
        self._config_file = config_file
        self._id = id
        self._version = version
        self._description = description
        self._created_at = created_at
        self._creator = creator

        self._inference_logger = util.get_obj_from_json(
            inference_logger, InferenceLogger
        )
        self._transformer = util.get_obj_from_json(transformer, Transformer)
        self._validate_script_file(self._model_framework, self._script_file)
        self._api_protocol = api_protocol
        self._environment = environment
        self._project_namespace = project_namespace
        self._project_name = None

    @public
    def deploy(self) -> deployment.Deployment:
        """Create a deployment for this predictor and persists it in the Model Serving.

        Returns:
            The deployment metadata object of a new or existing deployment.

        Examples:
            ```python

            import hopsworks

            project = hopsworks.login()

            # get Hopsworks Model Registry handle
            mr = project.get_model_registry()

            # retrieve the trained model you want to deploy
            my_model = mr.get_model("my_model", version=1)

            # get Hopsworks Model Serving handle
            ms = project.get_model_serving()

            my_predictor = ms.create_predictor(my_model)
            my_deployment = my_predictor.deploy()

            print(my_deployment.get_state())
            ```
        """
        _deployment = deployment.Deployment(
            predictor=self, name=self._name, description=self._description
        )
        _deployment.save()

        return _deployment

    @public
    def describe(self):
        """Print a JSON description of the predictor."""
        util.pretty_print(self)

    def _set_state(self, state: PredictorState):
        """Set the state of the predictor."""
        self._state = state

    @classmethod
    def _validate_serving_tool(cls, serving_tool):
        if serving_tool is not None:
            if client.is_saas_connection():
                # only kserve supported in saasy hopsworks
                if serving_tool != PREDICTOR.SERVING_TOOL_KSERVE:
                    raise ValueError(
                        "KServe deployments are the only supported in Serverless Hopsworks"
                    )
                return serving_tool
            # if not saas, check valid serving_tool
            serving_tools = list(util.get_members(PREDICTOR, prefix="SERVING_TOOL"))
            if serving_tool not in serving_tools:
                raise ValueError(
                    "Serving tool '{}' is not valid. Possible values are '{}'".format(
                        serving_tool, ", ".join(serving_tools)
                    )
                )
        return serving_tool

    @classmethod
    def _validate_script_file(cls, model_framework, script_file):
        if script_file is None and (model_framework == MODEL.FRAMEWORK_PYTHON):
            raise ValueError(
                "Predictor scripts are required in deployments for custom Python models."
            )

    @classmethod
    def _infer_model_server(cls, model_framework):
        if model_framework == MODEL.FRAMEWORK_TENSORFLOW:
            return PREDICTOR.MODEL_SERVER_TF_SERVING
        if model_framework == MODEL.FRAMEWORK_LLM:
            return PREDICTOR.MODEL_SERVER_VLLM
        return PREDICTOR.MODEL_SERVER_PYTHON

    @classmethod
    def _get_default_serving_tool(cls):
        # set kserve as default if it is available
        return (
            PREDICTOR.SERVING_TOOL_KSERVE
            if client.is_kserve_installed()
            else PREDICTOR.SERVING_TOOL_DEFAULT
        )

    @classmethod
    def _validate_resources(cls, resources, serving_tool):
        if (
            resources is not None
            and serving_tool == PREDICTOR.SERVING_TOOL_KSERVE
            and cls._get_raw_num_instances(resources) != 0
            and client.is_scale_to_zero_required()
        ):
            # ensure scale-to-zero for kserve deployments when required
            raise ValueError(
                "Scale-to-zero is required for KServe deployments in this cluster. Please, set the number of instances to 0."
            )
        return resources

    @classmethod
    def _get_default_resources(cls, serving_tool):
        num_instances = (
            0  # enable scale-to-zero by default if required
            if serving_tool == PREDICTOR.SERVING_TOOL_KSERVE
            and client.is_scale_to_zero_required()
            else SCALING_CONFIG.MIN_NUM_INSTANCES
        )
        return PredictorResources(num_instances)

    @classmethod
    def for_model(cls, model, **kwargs):
        kwargs["model_name"] = model.name
        kwargs["model_path"] = model.model_path
        kwargs["model_version"] = model.version

        # get predictor for specific model, includes model type-related validations
        return util.get_predictor_for_model(model=model, **kwargs)

    @public
    @classmethod
    def for_server(cls, name: str, script_file: str, **kwargs):
        # get predictor for a HTTP server without model
        return util.get_predictor_for_server(
            name=name, script_file=script_file, **kwargs
        )

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if isinstance(json_decamelized, list):
            if len(json_decamelized) == 0:
                return []
            return [cls.from_json(predictor) for predictor in json_decamelized]
        if "count" in json_decamelized:
            if json_decamelized["count"] == 0:
                return []
            return [cls.from_json(predictor) for predictor in json_decamelized["items"]]
        return cls.from_json(json_decamelized)

    @classmethod
    def from_json(cls, json_decamelized):
        predictor = Predictor(**cls.extract_fields_from_json(json_decamelized))
        predictor._set_state(PredictorState.from_response_json(json_decamelized))
        return predictor

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        kwargs = {}
        kwargs["name"] = json_decamelized.pop("name")
        kwargs["description"] = util.extract_field_from_json(
            json_decamelized, "description"
        )
        kwargs["version"] = json_decamelized.pop("version")
        with_model = "model_version" in json_decamelized
        kwargs["model_name"] = util.extract_field_from_json(
            json_decamelized,
            "model_name",
            default=(kwargs["name"] if with_model else None),
        )
        kwargs["model_version"] = util.extract_field_from_json(
            json_decamelized, "model_version"
        )
        kwargs["model_path"] = util.extract_field_from_json(
            json_decamelized, "model_path"
        )
        kwargs["model_framework"] = (
            json_decamelized.pop("model_framework")
            if "model_framework" in json_decamelized
            else (
                MODEL.FRAMEWORK_SKLEARN if with_model else None
            )  # backward compatibility
        )
        kwargs["model_server"] = json_decamelized.pop("model_server")
        kwargs["serving_tool"] = json_decamelized.pop("serving_tool")
        kwargs["script_file"] = util.extract_field_from_json(
            json_decamelized, "predictor"
        )
        kwargs["config_file"] = util.extract_field_from_json(
            json_decamelized, "config_file"
        )
        kwargs["resources"] = PredictorResources.from_json(json_decamelized)
        kwargs["inference_logger"] = InferenceLogger.from_json(json_decamelized)
        kwargs["inference_batcher"] = InferenceBatcher.from_json(json_decamelized)
        kwargs["transformer"] = Transformer.from_json(json_decamelized)
        kwargs["id"] = json_decamelized.pop("id")
        kwargs["created_at"] = json_decamelized.pop("created")
        kwargs["creator"] = json_decamelized.pop("creator")
        kwargs["api_protocol"] = json_decamelized.pop("api_protocol")
        if "environment_dto" in json_decamelized:
            environment = json_decamelized.pop("environment_dto")
            kwargs["environment"] = environment["name"]
        kwargs["project_namespace"] = json_decamelized.pop("project_namespace")
        kwargs["scaling_configuration"] = PredictorScalingConfig.from_json(
            json_decamelized
        )
        return kwargs

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**self.extract_fields_from_json(json_decamelized))
        self._set_state(PredictorState.from_response_json(json_decamelized))
        return self

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        json = {
            "id": self._id,
            "name": self._name,
            "description": self._description,
            "version": self._version,
            "created": self._created_at,
            "creator": self._creator,
            "modelServer": self._model_server,
            "servingTool": self._serving_tool,
            "predictor": self._script_file,
            "configFile": self._config_file,
            "apiProtocol": self._api_protocol,
            "projectNamespace": self._project_namespace,
        }
        if self.model_name is not None:
            json = {**json, "modelName": self._model_name}
        if self.model_path is not None:
            json = {**json, "modelPath": self._model_path}
        if self.model_version is not None:
            json = {**json, "modelVersion": self._model_version}
        if self.model_framework is not None:
            json = {**json, "modelFramework": self._model_framework}
        if self.environment is not None:
            json = {**json, "environmentDTO": {"name": self._environment}}
        if self._resources is not None:
            json = {**json, **self._resources.to_dict()}
        if self._inference_logger is not None:
            json = {**json, **self._inference_logger.to_dict()}
        if self._inference_batcher is not None:
            json = {**json, **self._inference_batcher.to_dict()}
        if self._transformer is not None:
            json = {**json, **self._transformer.to_dict()}
        if self._scaling_configuration is not None:
            json = {**json, **self._scaling_configuration.to_dict()}
        return json

    @public
    @property
    def id(self):
        """Id of the predictor."""
        return self._id

    @public
    @property
    def name(self):
        """Name of the predictor."""
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @public
    @property
    def version(self):
        """Version of the predictor."""
        return self._version

    @public
    @property
    def description(self):
        """Description of the predictor."""
        return self._description

    @description.setter
    def description(self, description: str):
        self._description = description

    @public
    @property
    def model_name(self):
        """Name of the model deployed by the predictor."""
        return self._model_name

    @model_name.setter
    def model_name(self, model_name: str):
        self._model_name = model_name

    @public
    @property
    def model_path(self):
        """Model path deployed by the predictor."""
        return self._model_path

    @model_path.setter
    def model_path(self, model_path: str):
        self._model_path = model_path

    @public
    @property
    def model_version(self):
        """Model version deployed by the predictor."""
        return self._model_version

    @model_version.setter
    def model_version(self, model_version: int):
        self._model_version = model_version

    @public
    @property
    def model_framework(self):
        """Model framework of the model to be deployed by the predictor."""
        return self._model_framework

    @model_framework.setter
    def model_framework(self, model_framework: str):
        self._model_framework = model_framework
        self._model_server = self._infer_model_server(model_framework)

    @public
    @property
    def artifact_version(self):
        """Artifact version deployed by the predictor.

        Warning: Deprecated
            Artifact versions are deprecated in favor of deployment versions.
        """
        return self._version

    @artifact_version.setter
    def artifact_version(self, artifact_version: int | str):
        pass  # do nothing, kept for backward compatibility

    @public
    @property
    def artifact_files_path(self):
        """Path of the artifact files deployed by the predictor."""
        # "/Projects/{project_name}/Deployments/{name}/{version}"
        return "{}/{}/{}/{}/{}".format(
            "/Projects",
            self._project_name,
            MODEL_SERVING.DEPLOYMENTS_DATASET,
            str(self._name),
            str(self._version),
        )

    @public
    @property
    def artifact_path(self):
        """Path of the model artifact deployed by the predictor. Resolves to /Projects/{project_name}/Models/{name}/{version}/Artifacts/{artifact_version}/{name}_{version}_{artifact_version}.zip."""
        # TODO: Deprecated
        artifact_name = f"{self._model_name}_{str(self._model_version)}_{str(self._artifact_version)}.zip"
        return f"{self._model_path}/{str(self._model_version)}/Artifacts/{str(self._artifact_version)}/{artifact_name}"

    @public
    @property
    def model_server(self):
        """Model server used by the predictor."""
        return self._model_server

    @public
    @property
    def serving_tool(self):
        """Serving tool used to run the model server."""
        return self._serving_tool

    @serving_tool.setter
    def serving_tool(self, serving_tool: str):
        self._serving_tool = serving_tool

    @public
    @property
    def script_file(self):
        """Script file used to load and run the model."""
        return self._script_file

    @script_file.setter
    def script_file(self, script_file: str):
        self._script_file = script_file

    @public
    @property
    def config_file(self):
        """Model server configuration file passed to the model deployment.

        It can be accessed via `CONFIG_FILE_PATH` environment variable from a predictor or transformer script.
        For LLM deployments without a predictor script, this file is used to configure the vLLM engine.
        """
        return self._config_file

    @config_file.setter
    def config_file(self, config_file: str):
        self._config_file = config_file

    @public
    @property
    def inference_logger(self):
        """Configuration of the inference logger attached to this predictor."""
        return self._inference_logger

    @inference_logger.setter
    def inference_logger(self, inference_logger: InferenceLogger):
        self._inference_logger = inference_logger

    @public
    @property
    def transformer(self):
        """Transformer configuration attached to the predictor."""
        return self._transformer

    @transformer.setter
    def transformer(self, transformer: Transformer):
        self._transformer = transformer

    @public
    @property
    def created_at(self):
        """Created at date of the predictor."""
        return self._created_at

    @public
    @property
    def creator(self):
        """Creator of the predictor."""
        return self._creator

    @public
    @property
    def requested_instances(self):
        """Total number of requested instances in the predictor."""
        num_instances = self._get_raw_num_instances(self._resources)
        if self._transformer is not None:
            num_instances += self._get_raw_num_instances(self._transformer.resources)
        return num_instances

    @public
    @property
    def api_protocol(self):
        """API protocol enabled in the predictor (e.g., HTTP or GRPC)."""
        return self._api_protocol

    @api_protocol.setter
    def api_protocol(self, api_protocol):
        self._api_protocol = api_protocol

    @public
    @property
    def environment(self):
        """Name of the inference environment."""
        return self._environment

    @environment.setter
    def environment(self, environment):
        self._environment = environment

    @public
    @property
    def project_namespace(self):
        """Kubernetes project namespace."""
        return self._project_namespace

    @project_namespace.setter
    def project_namespace(self, project_namespace):
        self._project_namespace = project_namespace

    @public
    @property
    def project_name(self):
        """Name of the project the deployment belongs to."""
        return self._project_name

    @project_name.setter
    def project_name(self, project_name: str):
        self._project_name = project_name

    @public
    def get_endpoint_url(self) -> str | None:
        """Get the base endpoint URL for this predictor.

        Returns the base URL that can be used with external HTTP clients.
        This is the path-based routing base endpoint without any protocol-specific
        suffixes like `:predict` or `/v1`.

        If Istio client is not available, returns `None` (Hopsworks REST API
        doesn't support base-only endpoints).

        Returns:
            Base endpoint URL, or `None` if unavailable.

        Examples:
            ```python
            url = predictor.get_endpoint_url()
            # url = "https://host:port/v1/project/name"
            ```
        """
        from hsml.core import serving_api

        serving = serving_api.ServingApi()

        istio_client = client.istio.get_instance()
        if istio_client is not None:
            path_parts = serving._get_istio_inference_path(self, base_only=True)
            return f"{istio_client._base_url}/{'/'.join(str(p) for p in path_parts)}"

        # Hopsworks REST API doesn't support base-only endpoints
        return None

    @public
    def get_openai_url(self) -> str | None:
        """Get the OpenAI-compatible API URL for vLLM deployments.

        Returns the URL for OpenAI-compatible API endpoints (e.g., /v1/chat/completions).
        This method only returns a URL for LLM (vLLM) deployments.

        Returns:
            OpenAI-compatible URL (base URL + "/v1"), or `None` if not a LLM deployment.

        Examples:
            ```python
            url = predictor.get_openai_compatible_url()
            # url = "https://host:port/v1/project/name/v1"
            # Then use: url + "/chat/completions"
            ```
        """
        if self._model_server != PREDICTOR.MODEL_SERVER_VLLM:
            return None

        base_url = self.get_endpoint_url()
        if base_url is None:
            return None

        return f"{base_url}/v1"

    @public
    def get_inference_url(self) -> str | None:
        """Get the KServe inference URL for standard model deployments.

        Returns the full URL with `:predict` suffix for KServe inference protocol.
        This method only returns a URL for standard model deployments (non-vLLM,
        with a model attached).

        If Istio client is not available, falls back to Hopsworks REST API path.

        Returns:
            Inference URL with `:predict` suffix, or `None` if not a standard model deployment.

        Examples:
            ```python
            url = predictor.get_inference_url()
            # url = "https://host:port/v1/project/name/v1/models/name:predict"
            ```
        """
        from hsml.core import serving_api

        # Only for standard model deployments (has model, not vLLM)
        has_model = self._model_name is not None and self._model_version is not None
        if not has_model or self._model_server == PREDICTOR.MODEL_SERVER_VLLM:
            return None

        serving = serving_api.ServingApi()

        # Try Istio client first
        istio_client = client.istio.get_instance()
        if istio_client is not None:
            path_parts = serving._get_istio_inference_path(self, base_only=False)
            return f"{istio_client._base_url}/{'/'.join(str(p) for p in path_parts)}"

        # Fallback to Hopsworks REST API path
        hopsworks_client = client.get_instance()
        path_parts = serving._get_hopsworks_inference_path(
            hopsworks_client._project_id, self
        )
        return f"{hopsworks_client._base_url}/hopsworks-api/api/{'/'.join(str(p) for p in path_parts)}"

    def __repr__(self):
        desc = (
            f", description: {self._description!r}"
            if self._description is not None
            else ""
        )
        return f"Predictor(name: {self._name!r}" + desc + ")"



#
#   Copyright 2026 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

import humps
from hopsworks_common import util


if TYPE_CHECKING:
    from hsfs.core.job import Job


class Chart:
    """Metadata object used to provide Chart information."""

    def __init__(
        self,
        id: int | None = None,
        title: str | None = None,
        description: str | None = None,
        url: str | None = None,
        job: Job | None = None,
        width: int | None = None,
        height: int | None = None,
        x: int | None = None,
        y: int | None = None,
        **kwargs,
    ):
        self._id = id
        self._title = title
        self._description = description
        self._url = url
        self._job = job
        self._width = width
        self._height = height
        self._x = x
        self._y = y

    @classmethod
    def from_response_json(cls, json_dict: dict[str, Any]) -> list[Chart]:
        if json_dict is None:
            return None

        json_decamelized = humps.decamelize(json_dict)

        if isinstance(json_decamelized, list):
            return [cls(**item) for item in json_decamelized]
        return cls(**json_decamelized)

    def to_dict(self):
        return {
            "id": self._id,
            "title": self._title,
            "description": self._description,
            "url": self._url,
            "job": self._job,
            "width": self._width,
            "height": self._height,
            "x": self._x,
            "y": self._y,
        }

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    @property
    def id(self) -> int | None:
        return self._id

    @id.setter
    def id(self, id: int) -> None:
        self._id = id

    @property
    def title(self) -> str | None:
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        self._title = title

    @property
    def description(self) -> str | None:
        return self._description

    @description.setter
    def description(self, description: str) -> None:
        self._description = description

    @property
    def url(self) -> str | None:
        return self._url

    @url.setter
    def url(self, url: str) -> None:
        self._url = url

    @property
    def job(self) -> Job | None:
        return self._job

    @job.setter
    def job(self, job: Job) -> None:
        self._job = job

    @property
    def width(self) -> int | None:
        return self._width

    @width.setter
    def width(self, width: int) -> None:
        self._width = width

    @property
    def height(self) -> int | None:
        return self._height

    @height.setter
    def height(self, height: int) -> None:
        self._height = height

    @property
    def x(self) -> int | None:
        return self._x

    @x.setter
    def x(self, x: int) -> None:
        self._x = x

    @property
    def y(self) -> int | None:
        return self._y

    @y.setter
    def y(self, y: int) -> None:
        self._y = y

    def delete(self) -> None:
        """Delete the chart from the feature store.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        from hsfs.core.chart_api import ChartApi

        ChartApi().delete_chart(self.id)

    def update(self) -> None:
        """Update the chart in the feature store.

        Updates the chart metadata with the current values of this object.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        from hsfs.core.chart_api import ChartApi

        ChartApi().update_chart(self)



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

from __future__ import annotations

from typing import TYPE_CHECKING

import hopsworks_common.client.istio.grpc.inference_client as inference_client
import hopsworks_common.client.istio.grpc.proto.grpc_predict_v2_pb2 as grpc__predict__v2__pb2


if TYPE_CHECKING:
    import grpc._channel


class GRPCInferenceServiceStub:
    """Inference Server GRPC endpoints."""

    def __init__(self, channel: grpc._channel.Channel):
        """Constructor.

        Parameters:
            channel: A grpc.Channel.
        """
        self.ServerLive = channel.unary_unary(
            "/inference.GRPCInferenceService/ServerLive",
            request_serializer=grpc__predict__v2__pb2.ServerLiveRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ServerLiveResponse.FromString,
        )
        self.ServerReady = channel.unary_unary(
            "/inference.GRPCInferenceService/ServerReady",
            request_serializer=grpc__predict__v2__pb2.ServerReadyRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ServerReadyResponse.FromString,
        )
        self.ModelReady = channel.unary_unary(
            "/inference.GRPCInferenceService/ModelReady",
            request_serializer=grpc__predict__v2__pb2.ModelReadyRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ModelReadyResponse.FromString,
        )
        self.ServerMetadata = channel.unary_unary(
            "/inference.GRPCInferenceService/ServerMetadata",
            request_serializer=grpc__predict__v2__pb2.ServerMetadataRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ServerMetadataResponse.FromString,
        )
        self.ModelMetadata = channel.unary_unary(
            "/inference.GRPCInferenceService/ModelMetadata",
            request_serializer=grpc__predict__v2__pb2.ModelMetadataRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ModelMetadataResponse.FromString,
        )
        self.ModelInfer = channel.unary_unary(
            "/inference.GRPCInferenceService/ModelInfer",
            request_serializer=grpc__predict__v2__pb2.ModelInferRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.ModelInferResponse.FromString,
        )
        self.RepositoryModelLoad = channel.unary_unary(
            "/inference.GRPCInferenceService/RepositoryModelLoad",
            request_serializer=grpc__predict__v2__pb2.RepositoryModelLoadRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.RepositoryModelLoadResponse.FromString,
        )
        self.RepositoryModelUnload = channel.unary_unary(
            "/inference.GRPCInferenceService/RepositoryModelUnload",
            request_serializer=grpc__predict__v2__pb2.RepositoryModelUnloadRequest.SerializeToString,
            response_deserializer=grpc__predict__v2__pb2.RepositoryModelUnloadResponse.FromString,
        )


class GRPCInferenceServiceServicer:
    """Inference Server GRPC endpoints."""

    def ServerLive(self, request, context):
        """The ServerLive API indicates if the inference server is able to receive and respond to metadata and inference requests.

        Parameters:
            request: The request message.
            context: The gRPC context.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ServerReady(self, request, context):
        """The ServerReady API indicates if the server is ready for inferencing.

        Parameters:
            request: The request message.
            context: The gRPC context.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ModelReady(self, request, context):
        """The ModelReady API indicates if a specific model is ready for inferencing.

        Parameters:
            request: The request message.
            context: The gRPC context.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ServerMetadata(self, request, context):
        """The ServerMetadata API provides information about the server.

        Parameters:
            request: The request message.
            context: The gRPC context.

        Errors are indicated by the google.rpc.Status returned for the request.
        The OK code indicates success and other codes indicate failure.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ModelMetadata(self, request, context):
        """The per-model metadata API provides information about a model.

        Parameters:
            request: The request message.
            context: The gRPC context.

        Errors are indicated by the google.rpc.Status returned for the request.
        The OK code indicates success and other codes indicate failure.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ModelInfer(self, request, context):
        """The ModelInfer API performs inference using the specified model.

        Parameters:
            request: The request message.
            context: The gRPC context.

        Errors are indicated by the google.rpc.Status returned for the request.
        The OK code indicates success and other codes indicate failure.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RepositoryModelLoad(self, request, context):
        """Load or reload a model from a repository.

        Parameters:
            request: The request message.
            context: The gRPC context.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RepositoryModelUnload(self, request, context):
        """Unload a model.

        Parameters:
            request: The request message.
            context: The gRPC context.
        """
        context.set_code(inference_client.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_GRPCInferenceServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "ServerLive": inference_client.unary_unary_rpc_method_handler(
            servicer.ServerLive,
            request_deserializer=grpc__predict__v2__pb2.ServerLiveRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ServerLiveResponse.SerializeToString,
        ),
        "ServerReady": inference_client.unary_unary_rpc_method_handler(
            servicer.ServerReady,
            request_deserializer=grpc__predict__v2__pb2.ServerReadyRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ServerReadyResponse.SerializeToString,
        ),
        "ModelReady": inference_client.unary_unary_rpc_method_handler(
            servicer.ModelReady,
            request_deserializer=grpc__predict__v2__pb2.ModelReadyRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ModelReadyResponse.SerializeToString,
        ),
        "ServerMetadata": inference_client.unary_unary_rpc_method_handler(
            servicer.ServerMetadata,
            request_deserializer=grpc__predict__v2__pb2.ServerMetadataRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ServerMetadataResponse.SerializeToString,
        ),
        "ModelMetadata": inference_client.unary_unary_rpc_method_handler(
            servicer.ModelMetadata,
            request_deserializer=grpc__predict__v2__pb2.ModelMetadataRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ModelMetadataResponse.SerializeToString,
        ),
        "ModelInfer": inference_client.unary_unary_rpc_method_handler(
            servicer.ModelInfer,
            request_deserializer=grpc__predict__v2__pb2.ModelInferRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.ModelInferResponse.SerializeToString,
        ),
        "RepositoryModelLoad": inference_client.unary_unary_rpc_method_handler(
            servicer.RepositoryModelLoad,
            request_deserializer=grpc__predict__v2__pb2.RepositoryModelLoadRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.RepositoryModelLoadResponse.SerializeToString,
        ),
        "RepositoryModelUnload": inference_client.unary_unary_rpc_method_handler(
            servicer.RepositoryModelUnload,
            request_deserializer=grpc__predict__v2__pb2.RepositoryModelUnloadRequest.FromString,
            response_serializer=grpc__predict__v2__pb2.RepositoryModelUnloadResponse.SerializeToString,
        ),
    }
    generic_handler = inference_client.method_handlers_generic_handler(
        "inference.GRPCInferenceService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class GRPCInferenceService:
    """Inference Server GRPC endpoints."""

    @staticmethod
    def ServerLive(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ServerLive",
            grpc__predict__v2__pb2.ServerLiveRequest.SerializeToString,
            grpc__predict__v2__pb2.ServerLiveResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ServerReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ServerReady",
            grpc__predict__v2__pb2.ServerReadyRequest.SerializeToString,
            grpc__predict__v2__pb2.ServerReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ModelReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ModelReady",
            grpc__predict__v2__pb2.ModelReadyRequest.SerializeToString,
            grpc__predict__v2__pb2.ModelReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ServerMetadata(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ServerMetadata",
            grpc__predict__v2__pb2.ServerMetadataRequest.SerializeToString,
            grpc__predict__v2__pb2.ServerMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ModelMetadata(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ModelMetadata",
            grpc__predict__v2__pb2.ModelMetadataRequest.SerializeToString,
            grpc__predict__v2__pb2.ModelMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ModelInfer(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ModelInfer",
            grpc__predict__v2__pb2.ModelInferRequest.SerializeToString,
            grpc__predict__v2__pb2.ModelInferResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RepositoryModelLoad(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/RepositoryModelLoad",
            grpc__predict__v2__pb2.RepositoryModelLoadRequest.SerializeToString,
            grpc__predict__v2__pb2.RepositoryModelLoadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RepositoryModelUnload(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return inference_client.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/RepositoryModelUnload",
            grpc__predict__v2__pb2.RepositoryModelUnloadRequest.SerializeToString,
            grpc__predict__v2__pb2.RepositoryModelUnloadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import builtins
from typing import TYPE_CHECKING, Literal

import hopsworks


if TYPE_CHECKING:
    from hopsworks_common.project import Project


def login(
    host: str | None = None,
    port: int = 443,
    project: str | None = None,
    api_key_value: str | None = None,
    api_key_file: str | None = None,
    hostname_verification: bool = False,
    trust_store_path: str | None = None,
    engine: Literal["spark", "python", "training", "spark-no-metastore", "spark-delta"]
    | None = "python",
) -> Project:
    """Login to Hopsworks and return the project.

    This function wraps hopsworks.login to simulate user input by overriding the built-in input function.

    Parameters:
        host: Hopsworks host URL.
        port: Hopsworks port (default 443).
        project: Project name to access.
        api_key_value: API key value for Hopsworks authentication.
        api_key_file: Path to a file containing the API key for Hopsworks authentication.
        hostname_verification: Enable hostname verification for Hopsworks authentication.
        trust_store_path: Path to the trust store for Hopsworks authentication.
        engine: Engine to use (default: python).

    Returns:
        Project: The project object after successful login.
    """
    # Backup original input
    original_input = builtins.input

    # Override input to simulate user input. Choose the first project by default.
    builtins.input = lambda _: 1

    try:
        project = hopsworks.login(
            host=host,
            port=port,
            api_key_value=api_key_value,
            api_key_file=api_key_file,
            hostname_verification=hostname_verification,
            trust_store_path=trust_store_path,
            engine=engine,
        )
    finally:
        # Restore original input function
        builtins.input = original_input

    return project



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

from abc import abstractmethod

from hopsworks_apigen import also_available_as
from hopsworks_common.client import base
from hopsworks_common.client.istio.grpc.inference_client import (
    GRPCInferenceServerClient,
)


@also_available_as("hsml.client.istio.base.Client")
class Client(base.Client):
    SERVING_API_KEY = "SERVING_API_KEY"
    HOPSWORKS_PUBLIC_HOST = "HOPSWORKS_PUBLIC_HOST"
    TOKEN_EXPIRED_MAX_RETRIES = 0

    BASE_PATH_PARAMS = []

    @abstractmethod
    def __init__(self):
        """To be implemented by clients."""

    def _get_verify(self, verify: str, trust_store_path: str | None) -> str | bool:
        """Get verification method for sending inference requests to Istio.

        Credit to https://gist.github.com/gdamjan/55a8b9eec6cf7b771f92021d93b87b2c

        Parameters:
            verify: Whether to perform hostname verification, is interpreted as `True` only if it is the string `"true"`.
            trust_store_path: The path of the truststore locally if it was uploaded manually to the external environment such as EKS or AKS.

        Returns:
            If verify is true and the truststore is provided, then return the trust store location; if verify is true but the truststore wasn't provided, then return true; if verify is false, then return false.
        """
        if verify == "true":
            if trust_store_path is not None:
                return trust_store_path
            return True

        return False

    def _get_host_port_pair(self) -> tuple[str, str]:
        """Removes "http or https" from the rest endpoint and returns a list [endpoint, port], where endpoint is on the format /path.. without http://.

        Returns:
            a tuple (endpoint, port)
        """
        endpoint = self._base_url
        if endpoint.startswith("http"):
            last_index = endpoint.rfind("/")
            endpoint = endpoint[last_index + 1 :]
        host, port = endpoint.split(":")
        return host, port

    def _close(self):
        """Closes a client. Can be implemented for clean up purposes, not mandatory."""
        self._connected = False

    def _create_grpc_channel(self, path_prefix: str) -> GRPCInferenceServerClient:
        return GRPCInferenceServerClient(
            url=self._host + ":" + str(self._port),
            path_prefix=path_prefix,
            serving_api_key=self._auth._token,
        )
